"""Pilatus implementation of fullCBF format, for use with Dectris detectors."""

from __future__ import annotations

import binascii
import re
import sys

from cctbx.eltbx import attenuation_coefficient

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFFull import FormatCBFFull
from dxtbx.format.FormatPilatusHelpers import determine_pilatus_mask
from dxtbx.format.FormatPilatusHelpers import get_vendortype as gv
from dxtbx.model import ParallaxCorrectedPxMmStrategy


class FormatCBFFullPilatus(FormatCBFFull):
    """An image reading class for full CBF format images from Pilatus
    detectors."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBFFull.get_cbf_header(image_file)

        for record in header.split("\n"):
            if "_array_data.header_convention" in record and "PILATUS" in record:
                return True

        return False

    def _beam(self):
        """Return a working beam instance. Override polarization to be 0.999."""

        beam = self._beam_factory.imgCIF_H(self._get_cbf_handle())
        beam.set_polarization_fraction(0.999)
        return beam

    def _detector(self):
        """Return a working detector instance, with added mask regions."""

        detector = self._detector_factory.imgCIF_H(self._get_cbf_handle(), "PAD")

        for f0, f1, s0, s1 in determine_pilatus_mask(detector):
            detector[0].add_mask(f0 - 1, s0 - 1, f1, s1)

        m = re.search(
            r"^#\s*(\S+)\ssensor, thickness\s*([0-9.]+)\s*m\s*$",
            self._cif_header,
            re.MULTILINE,
        )
        if m:
            # header gives thickness in metres, we store mm
            thickness = float(m.group(2)) * 1000
            material = m.group(1)

            if material == "Silicon":
                material = "Si"

            for panel in detector:
                panel.set_thickness(thickness)
                panel.set_material(material)

            try:
                # a header only CBF file will not have a beam object
                beam = self._beam()

            except Exception:
                beam = None

            if beam:
                # attenuation coefficient depends on the beam wavelength
                wavelength = beam.get_wavelength()

                # this will fail for undefined composite materials
                table = attenuation_coefficient.get_table(material)
                # mu_at_angstrom returns cm^-1
                mu = table.mu_at_angstrom(wavelength) / 10.0

                for panel in detector:
                    panel.set_px_mm_strategy(
                        ParallaxCorrectedPxMmStrategy(mu, thickness)
                    )
                    panel.set_mu(mu)

        m = re.search(r"^#\s*Detector:\s+(.*?)\s*$", self._cif_header, re.MULTILINE)
        if m and m.group(1):
            for panel in detector:
                panel.set_identifier(m.group(1).encode())

        size = detector[0].get_image_size()
        if size == (2463, 2527):
            self.vendortype = "Pilatus-6M"
        elif size == (1475, 1679):
            self.vendortype = "Pilatus-2M"
        elif size == (487, 619):
            self.vendortype = "Pilatus-300K"

        return detector

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values

    def get_raw_data(self):
        if self._raw_data is None:
            data = self._read_cbf_image()
            self._raw_data = data

        return self._raw_data

    def get_vendortype(self):
        return gv(self.get_detector())


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFFullPilatus.understand(arg))



from __future__ import annotations

from iotbx.detectors.dtrek import DTREKImage

from dxtbx.format.FormatSMVRigakuA200 import FormatSMVRigakuA200


class FormatSMVRigakuA200SPring8BL26B1(FormatSMVRigakuA200):
    """A class for reading SMV format Rigaku A200 images written by a detector
    usually at SPring-8 BL26B1, which requires a reversed axis goniometer"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a Rigaku A200 SMV format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMVRigakuA200SPring8BL26B1.get_smv_header(image_file)

        detector_prefix = header["DETECTOR_NAMES"].split()[0].strip()

        test = "%s%s" % (detector_prefix, "DETECTOR_IDENTIFICATION")
        if header.get(test) != "MSC_REIT_A200_SN09250183":
            return False

        test = "%s%s" % (detector_prefix, "DETECTOR_DESCRIPTION")
        if not header.get(test).startswith("A200"):
            return False

        test = "%s%s" % (detector_prefix, "GONIO_DESCRIPTION")
        if not header.get(test).startswith("SPring-8"):
            return False

        return True

    def detectorbase_start(self):
        self.detectorbase = DTREKImage(self._image_file)
        self.detectorbase.readHeader()

    def _goniometer(self):
        """Initialize the structure for the goniometer. Although ROTATION_VECTOR
        is written into the header, this might (?) always be 1. 1. 0. for this
        detector, and we want a reversed axis goniometer"""

        # axis = tuple(map(float, self._header_dictionary[
        #    'ROTATION_VECTOR'].split()))

        return self._goniometer_factory.single_axis_reverse()



from __future__ import annotations

import functools
import os
import sys
import time
from itertools import groupby

import numpy as np
from scipy.signal import convolve

import serialtbx.detector.xtc
import serialtbx.util
from libtbx import Auto
from libtbx.phil import parse
from scitbx.array_family import flex

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format, abstract
from dxtbx.format.FormatMultiImage import FormatMultiImage, Reader
from dxtbx.format.FormatStill import FormatStill
from dxtbx.model import Spectrum
from dxtbx.util.rotate_and_average import rotate_and_average

try:
    import psana
except ImportError:
    psana = None
except TypeError:
    # Check if SIT_* environment variables are set
    import os

    if os.environ.get("SIT_ROOT"):
        # Variables are present, so must have been another error
        raise
    psana = None

locator_str = """
  hits_file = None
    .type = str
    .help = path to a file where each line is 2 numbers separated by a space, a run index, and an event index in the XTC stream
  experiment = None
    .type = str
    .help = Experiment identifier, e.g. mfxo1916
  run = None
    .type = int
    .multiple = True
    .help = Run number or a list of runs to process
  mode = Auto
    .type = str
    .help = Mode for reading the xtc data (see LCLS documentation)
  data_source = None
    .type = str
    .help = Complete LCLS data source.  Overrides experiment and run.  Example: \
            exp=mfxo1916:run=20:smd \
            More info at https://confluence.slac.stanford.edu/display/PSDM/Manual#Manual-Datasetspecification
  detector_address = None
    .type = str
    .multiple = True
    .help = detector used for collecting the data at LCLS
  calib_dir = None
    .type = str
    .help = Specify path to custom calib directory if needed
  use_ffb = False
    .type = bool
    .help = Run on the ffb if possible. Only for active users!
  wavelength_delta_k = 0
    .type = float
    .help = Correction factor, needed during 2014
  wavelength_offset = None
    .type = float
    .help = Optional constant shift to apply to each wavelength. Note, if \
            spectrum_required = False and spectra calibration constants \
            (spectrum_eV_per_pixel and spectrum_eV_offset) are provided, \
            wavelength_offset can be used to apply a general correction \
            for any events with a dropped spectrum. If the spectrum is \
            present and calibration constants are provided, \
            wavelength_offset is ignored.
  wavelength_scale = None
    .type = float
    .help = Optional scalar to apply to each wavelength (see \
            wavelength_offset).  If both scale and offset are present, \
            the final wavelength is (scale * initial wavelength) + offset.
  wavelength_fallback = None
    .type = float
    .help = If the wavelength cannot be found from the XTC stream, fall \
            back to using this value instead
  spectrum_address = feespec
    .type = str
    .help = Address for incident beam spectrometer
  spectrum_eV_per_pixel = None
    .type = float
    .help = If not None, use the FEE spectrometer to determine the wavelength. \
            spectrum_eV_offset should be also specified. A weighted average of \
            the horizontal projection of the per-shot FEE spectrometer is used.\
            The equation for each pixel eV is \
            eV = (spectrum_eV_per_pixel * pixel_number) + spectrum_eV_offset
  spectrum_eV_offset = None
    .type = float
    .help = See spectrum_eV_per_pixel
  spectrum_rotation_angle = None
    .type = float
    .help = If present, first rotate the spectrum image a given amount in \
            degrees. Only applies to 2D spectrometers
  spectrum_pedestal = None
    .type = path
    .help = Path to pickled pedestal file to subtract from the pedestal
  spectrum_required = False
    .type = bool
    .help = Raise an exception for any event where the spectrum is not \
            available.
  spectrum_index_offset = None
    .type = int
    .help = Optional offset if the spectrometer and images are not in sync in \
            XTC stream
  check_spectrum {
    enable = False
      .type = bool
    smooth_window = 50
      .type = int
      .help = Before determining spectral width, smooth it by convolution with \
              a box of this width in pixels.
    max_width = .003
      .type = float
      .help = Reject spectra with greater than this fractional width.
    intensity_threshold = 0.2
      .type = float
      .help = Determine the spectral width at this fraction of the maximum.
    min_height = 500
      .type = float
      .help = Reject spectra below this max intensity (after smoothing).
  }
  filter {
    evr_address = evr1
      .type = str
      .help = Address for evr object which stores event codes. Should be evr0,\
              evr1, or evr2.
    required_present_codes = None
      .type = int
      .multiple = True
      .help = These codes must be present to keep the event
    required_absent_codes = None
      .type = int
      .multiple = True
      .help = These codes must be absent to keep the event
    pre_filter = False
      .type = bool
      .help = If True, read the event codes for all events up front, and \
              apply the filter then. Otherwise, apply when loading an \
              event.
  }
"""
locator_scope = parse(locator_str)


class XtcReader(Reader):
    def nullify_format_instance(self):
        """No-op for XTC streams. No issue with multiprocessing."""


@abstract
class FormatXTC(FormatMultiImage, FormatStill, Format):
    def __init__(self, image_file, **kwargs):
        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self.lazy = kwargs.get("lazy", True)
        FormatMultiImage.__init__(self, **kwargs)
        FormatStill.__init__(self, image_file, **kwargs)
        Format.__init__(self, image_file, **kwargs)
        self.current_index = None
        self.current_event = None
        self._psana_runs = {}  # empty container, to prevent breaking other formats
        if "locator_scope" in kwargs:
            self.params = FormatXTC.params_from_phil(
                master_phil=kwargs["locator_scope"], user_phil=image_file, strict=True
            )
        else:
            self.params = FormatXTC.params_from_phil(
                master_phil=locator_scope, user_phil=image_file, strict=True
            )

        if self.params.mode in [Auto, "Auto"]:
            xtc_version = getattr(psana, "xtc_version", None)
            if xtc_version == 2:
                self.params.mode = "psana2_idx"
            elif xtc_version is None:
                self.params.mode = "idx"
            else:
                assert False, "Unrecognized psana version"
        assert self.params.mode in [
            "idx",
            "smd",
            "psana2_idx",
        ], (
            "idx or smd mode should be used for LCLS-I analysis (idx is often faster). psana2_idx should be used for LCLS-II."
        )

        self._ds = FormatXTC._get_datasource(image_file, self.params)
        self._evr = None
        self._load_hit_indices()
        self.populate_events()

        self._cached_psana_detectors = {}
        self._beam_index = None
        self._beam_cache = None
        self._initialized = True
        self._fee = None

        if self.params.spectrum_pedestal:
            from libtbx import easy_pickle

            self._spectrum_pedestal = easy_pickle.load(self.params.spectrum_pedestal)
        else:
            self._spectrum_pedestal = None

        """
        # Prototype code for checking automatically determining the offst between the eBeam
        # and the spectrometer
        if self.params.spectrum_eV_per_pixel is not None and self.params.wavelength_offset is None:
            i = 0; count = 0
            all_fee_wav, all_eBeam_wav = [],[]
            while i < self.get_num_images() and count < 200:
                evt = self._get_event(i)
                spectrum = self.get_spectrum(i)
                eBeam_wav = serialtbx.detector.xtc.evt_wavelength(evt, delta_k=self.params.wavelength_delta_k)
                i += 1
                if spectrum is None: continue
                if eBeam_wav is None: continue
                all_fee_wav.append(spectrum.get_weighted_wavelength())
                all_eBeam_wav.append(eBeam_wav)
                count += 1
            if count >= 200:
                mean_eBeam_wav = sum(all_eBeam_wav)/len(all_eBeam_wav)
                mean_fee_wav = sum(all_fee_wav)/len(all_fee_wav)
                self.params.wavelength_offset = mean_fee_wav - mean_eBeam_wav
        """

    def _load_hit_indices(self):
        self._hit_inds = None
        if self.params.hits_file is not None:
            assert self.params.mode == "idx"
            hits = np.loadtxt(self.params.hits_file, int)
            hits = list(map(tuple, hits))
            key = lambda x: x[0]  # noqa: E731
            gb = groupby(sorted(hits, key=key), key=key)
            # dictionary where key is run number, and vals are indices of hits
            self._hit_inds = {r: [ind for _, ind in group] for r, group in gb}

    @staticmethod
    def understand(image_file):
        """Extracts the datasource and detector_address from the image_file and then feeds it to PSANA
        If PSANA fails to read it, then input may not be an xtc/smd file. If success, then OK.
        If detector_address is not provided, a command line promp will try to get the address
        from the user"""
        if not psana:
            return False
        try:
            params = FormatXTC.params_from_phil(locator_scope, image_file)
        except Exception:
            return False
        if params is None:
            return False

        try:
            FormatXTC._get_datasource(image_file, params)
        except Exception:
            return False
        return True

    @staticmethod
    def params_from_phil(master_phil, user_phil, strict=False):
        """Read the locator file"""
        try:
            user_input = parse(file_name=user_phil)
            working_phil, unused = master_phil.fetch(
                sources=[user_input], track_unused_definitions=True
            )
            unused_args = ["%s=%s" % (u.path, u.object.words[0].value) for u in unused]
            if len(unused_args) > 0 and strict:
                for unused_arg in unused_args:
                    print(unused_arg)
                print(
                    "Incorrect or unused parameter in locator file. Please check and retry"
                )
                return None
            params = working_phil.extract()
            return params
        except Exception:
            return None

    @classmethod
    def get_reader(cls):
        """
        Return a reader class
        """
        return functools.partial(XtcReader, cls)

    def populate_events(self):
        """Read the timestamps from the XTC stream.  Assumes the psana idx mode of reading data.
        Handles multiple LCLS runs by concatenating the timestamps from multiple runs together
        in a single list and creating a mapping."""
        if self.params.mode == "idx":
            if hasattr(self, "times") and len(self.times) > 0:
                return
        elif self.params.mode == "smd":
            if hasattr(self, "run_mapping") and self.run_mapping:
                return

        if not self._psana_runs:
            self._psana_runs = self._get_psana_runs(self._ds)

        self.times = []
        self.run_mapping = {}

        if self.params.mode == "idx":
            for run_num, run in self._psana_runs.items():
                times = run.times()
                if self._hit_inds is not None and run_num not in self._hit_inds:
                    continue
                if self._hit_inds is not None and run_num in self._hit_inds:
                    temp = []
                    for i_hit in self._hit_inds[run_num]:
                        temp.append(times[i_hit])
                    times = tuple(temp)
                if (
                    self.params.filter.required_present_codes
                    or self.params.filter.required_absent_codes
                ) and self.params.filter.pre_filter:
                    times = [t for t in times if self.filter_event(run.event(t))]

                self.run_mapping[run_num] = (
                    len(self.times),
                    len(self.times) + len(times),
                    run,
                )

                self.times.extend(times)
            self.n_images = len(self.times)

        elif self.params.mode == "psana2_idx":
            """
            ### Code snippet from Mona on psana idx behavior ###

                 with run.build_table() as success:
                    if success:
                        # Extract first 10 valid timestamps (those with non-empty offset info)
                        valid_ts = sorted(k for k in run._ts_table if run._ts_table[k])[:10]
                        assert len(valid_ts) == 10, "Expected at least 10 L1Accept events with offsets"

                        # Randomly pick 3 timestamps to test
                        sample_ts = random.sample(valid_ts, 3)
                        for ts in sample_ts:
                            evt = run.event(ts)
                            assert evt is not None
                            assert len(evt._dgrams) > 0

                            det = run.Detector('xppcspad')
            """
            FormatXTC._get_datasource(self._image_file, self.params)
            assert len(self._psana_runs.items()) == 1

            for run_num, run in self._psana_runs.items():
                with run.build_table() as success:
                    if success:
                        times = sorted(run._ts_table)
                        if self._hit_inds is not None and run_num not in self._hit_inds:
                            continue
                        if self._hit_inds is not None and run_num in self._hit_inds:
                            temp = []
                            for i_hit in self._hit_inds[run_num]:
                                temp.append(times[i_hit])
                            times = tuple(temp)
                        if (
                            self.params.filter.required_present_codes
                            or self.params.filter.required_absent_codes
                        ) and self.params.filter.pre_filter:
                            if not self._evr:
                                timing = run.Detector("timing")
                                self._evr = timing.raw
                            times = [
                                t for t in times if self.filter_event(run.event(t))
                            ]
                        self.run_mapping[run_num] = (
                            len(self.times),
                            len(self.times) + len(times),
                            run,
                        )
                        self.times.extend(times)
                    else:
                        self.run_mapping[run_num] = (1, 1, run)

                self.n_images = len(self.times)

        elif self.params.mode == "smd":
            self._ds = FormatXTC._get_datasource(self._image_file, self.params)
            for event in self._ds.events():
                run = event.run()
                if run not in self.run_mapping:
                    self.run_mapping[run] = []
                if (
                    self.params.filter.required_present_codes
                    or self.params.filter.required_absent_codes
                ) and self.params.filter.pre_filter:
                    if self.filter_event(event):
                        self.run_mapping[run].append(event)
                else:
                    self.run_mapping[run].append(event)
            total = 0
            remade_mapping = {}
            for run in sorted(self.run_mapping):
                start = total
                end = len(self.run_mapping[run]) + total
                total += len(self.run_mapping[run])
                events = self.run_mapping[run]
                remade_mapping[run] = start, end, run, events
            self.run_mapping = remade_mapping
            self.n_images = sum(
                [
                    self.run_mapping[r][1] - self.run_mapping[r][0]
                    for r in self.run_mapping
                ]
            )

    def filter_event(self, evt):
        """Return True to keep the event, False to reject it."""
        if not (
            self.params.filter.required_present_codes
            or self.params.filter.required_absent_codes
        ):
            return True
        if not self._evr:
            self._evr = psana.Detector(self.params.filter.evr_address)
        try:  # psana1
            codes = self._evr.eventCodes(evt)
        except Exception:  # psana2_idx
            codes = [i for i, val in enumerate(self._evr.eventcodes(evt)) if val != 0]

        if self.params.filter.required_present_codes and not all(
            c in codes for c in self.params.filter.required_present_codes
        ):
            return False
        if self.params.filter.required_absent_codes and any(
            c in codes for c in self.params.filter.required_absent_codes
        ):
            return False
        return True

    def get_run_from_index(self, index=None):
        """Look up the run number given an index"""
        if index is None:
            index = 0
        for run_number in self.run_mapping:
            start, stop, run = self.run_mapping[run_number][0:3]
            if index >= start and index < stop:
                return run
        raise IndexError("Index is not within bounds")

    def _get_event(self, index):
        """Retrieve a psana event given and index. This is the slow step for reading XTC streams,
        so implement a cache for the last read event."""
        if index == self.current_index:
            return self.current_event
        else:
            self.current_index = index
            if self.params.mode == "idx":
                evt = self.get_run_from_index(index).event(self.times[index])
            elif self.params.mode == "psana2_idx":
                run = self.get_run_from_index(index)
                if not self._evr:
                    timing = run.Detector("timing")
                    self._evr = timing.raw
                evt = run.event(self.times[index])
            elif self.params.mode == "smd":
                for run_number in self.run_mapping:
                    start, stop, run, events = self.run_mapping[run_number]
                    if index >= start and index < stop:
                        evt = events[index - start]
            if (
                (
                    self.params.filter.required_present_codes
                    or self.params.filter.required_absent_codes
                )
                and not self.params.filter.pre_filter
                and not self.filter_event(evt)
            ):
                evt = None
            self.current_event = evt
            return self.current_event

    @staticmethod
    def _get_datasource(image_file, params, **kwargs):
        """Construct a psana data source object given the locator parameters"""
        if params.calib_dir is not None:
            psana.setOption("psana.calib-dir", params.calib_dir)
        if params.data_source is None:
            if (
                params.experiment is None
                or params.run is None
                or params.mode is None
                or len(params.run) == 0
            ):
                return False
            img = "exp=%s:run=%s:%s" % (
                params.experiment,
                ",".join(["%d" % r for r in params.run]),
                params.mode,
            )

            if params.use_ffb:
                # as ffb is only at SLAC, ok to hardcode /reg/d here
                img += ":dir=/reg/d/ffb/%s/%s/xtc" % (
                    params.experiment[0:3],
                    params.experiment,
                )
        else:
            img = params.data_source
        try:
            src = psana.DataSource(img)
        except psana.datasource.InvalidDataSource:
            # psana2
            assert len(params.run) == 1
            src = psana.DataSource(exp=params.experiment, run=params.run[0], **kwargs)
        return src

    @staticmethod
    def _get_psana_runs(datasource):
        """
        Extracts the runs,
        These can only be extracted once,
        only call this method after datasource is set
        """
        # this is key,value = run_integer, psana.Run, e.g. {62: <psana.Run(@0x7fbd0e23c990)>}
        psana_runs = {}
        r = next(datasource.runs())
        try:
            psana_runs[r.runnum] = r
        except AttributeError:  # r doesn't have runnum: psana1
            psana_runs[r.run()] = r
        return psana_runs

    def _get_psana_detector(self, run):
        """Returns the psana detector for the given run"""
        if self.params.mode == "psana2_idx":
            run_num = run.runnum
            if run_num not in self._cached_psana_detectors:
                assert len(self.params.detector_address) == 1
                self._cached_psana_detectors[run_num] = run.Detector(
                    self.params.detector_address[0]
                )
            return self._cached_psana_detectors[run_num]
        else:
            if run.run() not in self._cached_psana_detectors:
                assert len(self.params.detector_address) == 1
                self._cached_psana_detectors[run.run()] = psana.Detector(
                    self.params.detector_address[0], run.env()
                )
            return self._cached_psana_detectors[run.run()]

    def get_psana_timestamp(self, index):
        """Get the cctbx.xfel style event timestamp given an index"""
        evt = self._get_event(index)
        if not evt:
            return None
        time = evt.get(psana.EventId).time()
        # fid = evt.get(psana.EventId).fiducials()

        sec = time[0]
        nsec = time[1]

        return serialtbx.util.time.timestamp((sec, nsec / 1e6))

    def get_num_images(self):
        return self.n_images

    def get_beam(self, index=None):
        return self._beam(index)

    def check_spectrum(self, spectrum):
        """Verify the spectrum is above a certain threshold"""
        xvals = spectrum.get_energies_eV()
        yvals = spectrum.get_weights()
        window = self.params.check_spectrum.smooth_window
        yvals = convolve(yvals, np.ones((window,)) / window, mode="same")
        ymax = max(yvals)
        if ymax < self.params.check_spectrum.min_height:
            return False
        threshold = ymax * self.params.check_spectrum.intensity_threshold
        indices = np.where(yvals > threshold)[0]
        width = xvals[indices[-1]] - xvals[indices[0]]
        frac_width = width / spectrum.get_weighted_energy_eV()
        return frac_width < self.params.check_spectrum.max_width

    def _beam(self, index=None):
        """Returns a simple model for the beam"""
        if index is None:
            index = 0
        if self._beam_index != index:
            self._beam_index = index
            evt = self._get_event(index)
            if not evt:
                self._beam_cache = None
                return None
            spectrum = self.get_spectrum(index)
            if spectrum:
                if self.params.check_spectrum.enable:
                    if not self.check_spectrum(spectrum):
                        return None
                wavelength = spectrum.get_weighted_wavelength()
            else:
                try:
                    wavelength = serialtbx.detector.xtc.evt_wavelength(
                        evt, delta_k=self.params.wavelength_delta_k
                    )
                except TypeError:  # no eBeam
                    wavelength = None
                if wavelength is None or wavelength <= 0:
                    wavelength = self.params.wavelength_fallback
                if wavelength is not None and wavelength > 0:
                    if self.params.wavelength_scale is not None:
                        wavelength *= self.params.wavelength_scale
                    if self.params.wavelength_offset is not None:
                        wavelength += self.params.wavelength_offset
            if wavelength is None:
                self._beam_cache = None
            else:
                self._beam_cache = self._beam_factory.simple(wavelength)
            try:
                s, nsec = evt.get(psana.EventId).time()
            except AttributeError:
                # psana2
                ts = evt.timestamp
                s = int(str(ts)[0:10])  # not elegant but works.
            evttime = time.gmtime(s)
            if (
                evttime.tm_year == 2020 and evttime.tm_mon >= 7
            ) or evttime.tm_year > 2020:
                if self._beam_cache is not None:
                    self._beam_cache.set_polarization_normal((1, 0, 0))

        if self._beam_cache.get_wavelength() < 0.1:
            from dxtbx.model.beam import Probe

            self._beam_cache.set_probe(Probe.electron)

        return self._beam_cache

    def get_spectrum(self, index=None):
        if index is None:
            index = 0
        spectrum = self._spectrum(index)
        if not spectrum and self.params.spectrum_required:
            raise RuntimeError("No spectrum in shot %d" % index)
        return spectrum

    def _spectrum(self, index=None):
        if index is None:
            index = 0
        if self.params.spectrum_index_offset:
            index += self.params.spectrum_index_offset
            if index < 0:
                return None
        if self.params.spectrum_eV_per_pixel is None:
            return None

        evt = self._get_event(index)
        if not evt:
            return None
        if self._fee is None:
            try:
                self._fee = psana.Detector(self.params.spectrum_address)
            except AttributeError:  # psana2
                self._fee = (
                    self.get_run_from_index(index)
                    .Detector(self.params.spectrum_address)
                    .raw
                )
        if self._fee is None:
            return None
        if hasattr(self._fee, "hproj"):
            # psana2
            y = self._fee.hproj(evt)
            if y is None or not len(y):
                return None
            if self.params.spectrum_pedestal:
                y = y - self._spectrum_pedestal.as_numpy_array()
            x = (
                self.params.spectrum_eV_per_pixel * np.array(range(len(y)))
            ) + self.params.spectrum_eV_offset
        else:
            try:
                fee = self._fee.get(evt)
                y = fee.hproj()
                if self.params.spectrum_pedestal:
                    y = y - self._spectrum_pedestal.as_numpy_array()

            except AttributeError:  # Handle older spectometers without the hproj method
                try:
                    img = self._fee.image(evt)
                except AttributeError:
                    return None
                if self.params.spectrum_rotation_angle is None:
                    x = (
                        self.params.spectrum_eV_per_pixel
                        * np.array(range(img.shape[1]))
                    ) + self.params.spectrum_eV_offset
                    y = img.mean(axis=0)  # Collapse 2D image to 1D trace
                else:
                    mask = img == 2**16 - 1
                    mask = np.invert(mask)

                    x, y = rotate_and_average(
                        img, self.params.spectrum_rotation_angle, deg=True, mask=mask
                    )
                    x = (
                        self.params.spectrum_eV_per_pixel * x
                    ) + self.params.spectrum_eV_offset

            else:
                x = (
                    self.params.spectrum_eV_per_pixel * np.array(range(len(y)))
                ) + self.params.spectrum_eV_offset
        try:
            # Apply vignetting here to x and y
            sp = Spectrum(flex.double(x), flex.double(y))
        except RuntimeError:
            return None
        return sp

    def get_goniometer(self, index=None):
        return None

    def get_scan(self, index=None):
        return None


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatXTC.understand(arg))



from __future__ import annotations

import pytest

from dxtbx.model import tof_helpers


def test_wavelength_from_tof():
    L0 = 8.3
    L1 = 0.24062553870557318
    tof = 0.0017464535727308615
    wavelength = tof_helpers.wavelength_from_tof(L0 + L1, tof)
    assert wavelength == pytest.approx(0.8089606203802903)

    L1 = 0.23021894342637692
    tof = 0.0016982786984656873
    wavelength = tof_helpers.wavelength_from_tof(L0 + L1, tof)
    assert wavelength == pytest.approx(0.7876056098518008)


def test_tof_from_wavelength():
    L0 = 8.3
    L1 = 0.24062553870557318
    wavelength = 0.8089606203802903
    tof = tof_helpers.tof_from_wavelength(L0 + L1, wavelength)
    assert tof == pytest.approx(0.0017464535727308615)

    L1 = 0.23021894342637692
    wavelength = 0.7876056098518008
    tof = tof_helpers.tof_from_wavelength(L0 + L1, wavelength)
    assert tof == pytest.approx(0.0016982786984656873)


def test_frame_to_tof_interpolator():
    frames = [i + 1 for i in range(9)]
    tof = [
        500.5,
        501.5010070800781,
        502.5040283203125,
        503.5090026855469,
        504.5159912109375,
        505.5249938964844,
        506.5360107421875,
        507.54901123046875,
        508.5639953613281,
    ]
    interpolator = tof_helpers.frame_to_tof_interpolator(frames, tof)

    assert interpolator(1) == pytest.approx(500.5)
    assert interpolator(9) == pytest.approx(508.56399536)
    assert interpolator(5.5) == pytest.approx(505.02024061)
    assert interpolator([1, 9, 5.5]) == pytest.approx(
        (500.5, 508.56399536, 505.02024061)
    )
    with pytest.raises(ValueError):
        _ = interpolator(-1)
    with pytest.raises(ValueError):
        _ = interpolator(0)
    with pytest.raises(ValueError):
        _ = interpolator(10)
    with pytest.raises(ValueError):
        _ = interpolator([-1, 2, 3])

    with pytest.raises(AssertionError):
        frames2 = frames + [10]  # frames and tof different length
        _ = tof_helpers.frame_to_tof_interpolator(frames2, tof)
    with pytest.raises(AssertionError):
        frames2 = frames[:]
        frames2[-1] = 5  # frames not strictly increasing
        _ = tof_helpers.frame_to_tof_interpolator(frames2, tof)
    with pytest.raises(AssertionError):
        frames2 = frames[:]  # frames not all > 0
        frames2[0] = -1
        _ = tof_helpers.frame_to_tof_interpolator(frames2, tof)
    with pytest.raises(AssertionError):
        tof2 = tof[:]
        tof2[-1] = 5  # tof not strictly increasing
        _ = tof_helpers.frame_to_tof_interpolator(frames, tof2)
    with pytest.raises(AssertionError):
        tof2 = tof[:]  # tof not all > 0
        tof2[0] = -1
        _ = tof_helpers.frame_to_tof_interpolator(frames, tof2)


def test_tof_to_frame_interpolator():
    frames = [i + 1 for i in range(9)]
    tof = [
        500.5,
        501.5010070800781,
        502.5040283203125,
        503.5090026855469,
        504.5159912109375,
        505.5249938964844,
        506.5360107421875,
        507.54901123046875,
        508.5639953613281,
    ]
    interpolator = tof_helpers.tof_to_frame_interpolator(tof, frames)

    assert interpolator(500.5) == pytest.approx(1)
    assert interpolator(508.56399536) == pytest.approx(9)
    assert interpolator(505.02024061) == pytest.approx(5.5)
    assert interpolator((500.5, 508.56399536, 505.02024061)) == pytest.approx(
        (1, 9, 5.5)
    )
    with pytest.raises(ValueError):
        _ = interpolator(-1)
    with pytest.raises(ValueError):
        _ = interpolator(500)
    with pytest.raises(ValueError):
        _ = interpolator(510)
    with pytest.raises(ValueError):
        _ = interpolator([-1, 500.5, 503])

    with pytest.raises(AssertionError):
        frames2 = frames + [10]  # frames and tof different length
        _ = tof_helpers.tof_to_frame_interpolator(tof, frames2)
    with pytest.raises(AssertionError):
        frames2 = frames[:]
        frames2[-1] = 5  # frames not strictly increasing
        _ = tof_helpers.tof_to_frame_interpolator(tof, frames2)
    with pytest.raises(AssertionError):
        frames2 = frames[:]  # frames not all > 0
        frames2[0] = -1
        _ = tof_helpers.tof_to_frame_interpolator(tof, frames2)
    with pytest.raises(AssertionError):
        tof2 = tof[:]
        tof2[-1] = 5  # tof not strictly increasing
        _ = tof_helpers.tof_to_frame_interpolator(tof2, frames)
    with pytest.raises(AssertionError):
        tof2 = tof[:]  # tof not all > 0
        tof2[0] = -1
        _ = tof_helpers.tof_to_frame_interpolator(tof2, frames)



"""Tests for dxbtx.format.FormatROD format classes."""

from __future__ import annotations

import pytest

from dxtbx.format.FormatROD import FormatROD, FormatROD_Arc
from dxtbx.model.experiment_list import ExperimentListFactory


def test_HyPix6000_TY6(dials_data):
    filename = (
        dials_data("image_examples") / "Hypix6000-monoclinic_lysozyme1_1_1.rodhypix.bz2"
    )

    assert FormatROD.understand(filename)
    expts = ExperimentListFactory.from_filenames([filename])
    assert len(expts) == 1

    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatROD

    gonio = imageset.get_goniometer()
    axes = gonio.get_names()
    assert list(axes) == ["PHI", "KAPPA=CHI", "OMEGA"]

    detector = imageset.get_detector()
    assert len(detector) == 1
    assert detector[0].get_gain() == 1.0
    assert detector[0].get_origin() == pytest.approx(
        (-44.6944, -36.9736, -23.2606), rel=1e-5
    )

    # TY6 decompression test
    img = imageset.get_raw_data(0)[0]
    assert img.focus() == (800, 775)
    assert img[123, 456] == 3


def test_EosS2_TY5(dials_data):
    filename = (
        dials_data("image_examples") / "EosS2-pre_SHR248e_CHCl3_pentane_1_1.img.bz2"
    )

    assert FormatROD.understand(filename)
    expts = ExperimentListFactory.from_filenames([filename])
    assert len(expts) == 1

    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatROD

    gonio = imageset.get_goniometer()
    axes = gonio.get_names()
    assert list(axes) == ["PHI", "KAPPA=CHI", "OMEGA"]

    detector = imageset.get_detector()
    assert len(detector) == 1
    # this CCD image has a non-unity gain
    assert detector[0].get_gain() == 100.0
    # if binning was not properly taken into account, the origin would break
    assert detector[0].get_origin() == pytest.approx(
        (-48.2605, -30.8897, -17.5460), rel=1e-5
    )

    # TY5 decompression test
    img = imageset.get_raw_data(0)[0]
    assert img.focus() == (512, 512)
    assert img[123, 456] == 174


def test_HyPixArc_TY6(dials_data):
    # HyPix-Arc is also tested in an end-to-end test `test_rigaku_hypix_arc_150()`
    # in dials/tests/algorithms/indexing/test_index.py.

    filename = dials_data("Rigaku_HyPix_Arc") / "CHNOS_1_2.rodhypix"

    assert FormatROD_Arc.understand(filename)
    expts = ExperimentListFactory.from_filenames([filename])
    assert len(expts) == 1

    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatROD_Arc

    # HyPix-Arc consists of three panels
    detector = imageset.get_detector()
    assert len(detector) == 3



"""An implementation of the CBF image reader for Eiger images"""

from __future__ import annotations

import datetime
import sys

from dxtbx.format.FormatCBFMiniEiger import FormatCBFMiniEiger


class FormatCBFMiniEigerPetraP14(FormatCBFMiniEiger):
    """A class for reading mini CBF format Eiger images, and correctly
    constructing a model for the experiment from this. This tuned for Petra P14"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Eiger mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniEiger.get_cbf_header(image_file)

        # Valid from 22nd May 2021
        expected_serial = "E-32-0129"
        if timestamp := FormatCBFMiniEiger._get_timestamp_from_raw_header(header):
            # We have a timestamp. Let's see what detector we should expect

            # Before 22nd May 2021
            if timestamp < datetime.datetime(2021, 5, 22):
                expected_serial = "E-32-0107"

        # Find the line recording detector serial, and check
        for record in header.split("\n"):
            if (
                "# detector" in record.lower()
                and "eiger" in record.lower()
                and expected_serial in record
            ):
                return True

        return False

    def _goniometer(self):
        return self._goniometer_factory.known_axis((0, 1, 0))


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniEigerPetraP14.understand(arg))



from __future__ import annotations

import argparse
import os

import dxtbx.format.Registry
import dxtbx.util

dag = dxtbx.format.Registry.get_format_class_dag()


def recurse(parentformat, filename):
    for subformat in dag.get(parentformat, []):
        understood = dxtbx.format.Registry.get_format_class_for(subformat).understand(
            filename
        )
        print(f"{subformat}: {understood}")
        if understood:
            recurse(subformat, filename)


def show_matching_formats(files):
    for filename in files:
        print(f"\n=== {filename} ===")
        if os.path.exists(filename):
            recurse("Format", filename)
        else:
            print("File not found.")


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser()
    parser.add_argument("filenames", metavar="IMAGE", nargs="+")
    options = parser.parse_args(args)
    show_matching_formats(options.filenames)


if __name__ == "__main__":
    run()



from __future__ import annotations

import glob

import h5py
import pytest

from libtbx.phil import parse
from libtbx.test_utils import approx_equal

from dxtbx.format.nexus import h5str
from dxtbx.model.experiment_list import ExperimentListFactory

pytest.importorskip("dials")

from dxtbx.format.nxmx_writer import NXmxWriter, phil_scope  # noqa: E402


def test_writer_jf16M(dials_data, tmpdir):
    h5path = (
        dials_data("lysozyme_JF16M_4img") / "lyso009a_0087.JF07T32V01_master_4img.h5"
    )
    with h5py.File(h5path) as handle:
        output_file = tmpdir / "4img.h5"
        instrument_name = h5str(handle["entry/instrument/name"][()])
        source_name = h5str(handle["entry/source/name"][()])
        start_time = h5str(handle["entry/start_time"][()])
        end_time_estimated = h5str(handle["entry/end_time_estimated"][()])
        sample_name = h5str(handle["entry/sample/name"][()])
        params = phil_scope.fetch(
            parse(
                f"""
        output_file = {output_file}
        nexus_details {{
          instrument_name = {instrument_name}
          source_name = {source_name}
          start_time = {start_time}
          end_time_estimated = {end_time_estimated}
          sample_name = {sample_name}
        }}
        """
            )
        ).extract()

    expts1 = ExperimentListFactory.from_filenames([h5path])

    NXmxWriter(params)(expts1)

    expts2 = ExperimentListFactory.from_filenames([output_file])

    def recursive_test(pg1, pg2):
        assert pg1.is_similar_to(pg2)
        for c1, c2 in zip(pg1, pg2):
            recursive_test(c1, c2)

    for d1, d2 in zip(expts1.detectors(), expts2.detectors()):
        recursive_test(d1.hierarchy(), d2.hierarchy())

    for b1, b2 in zip(expts1.beams(), expts2.beams()):
        assert approx_equal(b1.get_wavelength(), b2.get_wavelength())
        assert approx_equal(b2.get_s0(), b2.get_s0())

    assert not (any(expts2.scans()))
    assert not (any(expts2.goniometers()))


def test_writer_x4wide(dials_data, tmpdir):
    cbfspath = dials_data("x4wide") / "X4_wide_M1S4_2_000*.cbf"

    output_file = tmpdir / "x4wide_10img.h5"
    params = phil_scope.fetch(
        parse(
            f"""
    output_file = {output_file}
    nexus_details {{
      instrument_name = PILATUS
      source_name =  PILATUS 6M-F, S/N 60-0114-F
      start_time = 2013-02-08T13:23:40.833
      end_time_estimated = 2013-02-08T13:24:00.000
      sample_name = sample_x4wide
    }}
    """
        )
    ).extract()

    expts1 = ExperimentListFactory.from_filenames(glob.glob(str(cbfspath)))

    NXmxWriter(params)(expts1)

    expts2 = ExperimentListFactory.from_filenames([tmpdir / "x4wide_10img.h5"])

    def recursive_test(pg1, pg2):
        assert pg1.is_similar_to(pg2)
        for c1, c2 in zip(pg1, pg2):
            recursive_test(c1, c2)

    for d1, d2 in zip(expts1.detectors(), expts2.detectors()):
        recursive_test(d1.hierarchy(), d2.hierarchy())

    for b1, b2 in zip(expts1.beams(), expts2.beams()):
        assert approx_equal(b1.get_wavelength(), b2.get_wavelength())
        assert approx_equal(b2.get_s0(), b2.get_s0())

    for scan1, scan2 in zip(expts1.scans(), expts2.scans()):
        assert approx_equal(scan1.get_array_range(), scan2.get_array_range())
        assert approx_equal(scan1.get_oscillation(), scan2.get_oscillation())

    for gonio1, gonio2 in zip(expts1.goniometers(), expts2.goniometers()):
        assert approx_equal(gonio1.get_rotation_axis(), gonio2.get_rotation_axis())



from __future__ import annotations

import os
import sys

import h5py
import numpy as np

from cctbx import factor_ev_angstrom
from scitbx.array_family import flex

from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.format.FormatStill import FormatStill
from dxtbx.format.nexus import h5str


class FormatHDF5SaclaRayonix(FormatHDF5, FormatStill):
    """
    Class to handle multi-event HDF5 files from Rayonix detector (MX300-HS)
    preprocessed by Cheetah SFX pipeline at SACLA.
    """

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file, "r") as h5_handle:
            if "metadata/detector" not in h5_handle:
                return False
            if h5str(h5_handle["metadata/detector"][()]) != "Rayonix MX300HS":
                return False
            if any(elem.startswith("tag-") for elem in h5_handle):
                return True

        return False

    def __init__(self, image_file, index=0, reconst_mode=False, **kwargs):
        self._raw_data = None
        self.index = index
        self.image_filename = image_file
        super().__init__(image_file, **kwargs)

        # This hard-coded value can be overwritten
        # by RAYONIX_DISTANCE

        self.distance = 240.0  # mm
        self.mask = None

        # Read metadata if possible
        # Read pixel size from the metadata and determine the binning of rayonix
        self.read_metadata()
        self.PIXEL_SIZE = self.pixelsize_in_um / 1000  # convert um to mm
        self.bin_size = int(self.pixelsize_in_um / 39.1)
        self.RECONST_SIZE = 7680 // self.bin_size

        # Override by environmental variables
        if os.getenv("RAYONIX_DISTANCE"):
            self.distance = float(os.environ["RAYONIX_DISTANCE"])

    def _start(self):
        h5_handle = h5py.File(self.image_filename, "r")

        self._images = sorted([tag for tag in h5_handle if tag.startswith("tag-")])
        self.tag = self._images[self.index]
        h5_handle.close()

    def read_metadata(self):
        h5_handle = h5py.File(self.image_filename, "r")
        self.pixelsize_in_um = h5_handle["metadata"]["pixelsize_in_um"][()]
        h5_handle.close()

    def get_image_file(self, index=None):
        return self.image_filename

    def set_index(self, index):
        assert index < len(self._images)

        self.index = index
        self.tag = self._images[self.index]
        self._raw_data = None

    def _detector(self, index=None):
        return self._detector_factory.simple(
            sensor="CCD",
            distance=self.distance,
            beam_centre=(
                self.RECONST_SIZE / 2 * self.PIXEL_SIZE,
                self.RECONST_SIZE / 2 * self.PIXEL_SIZE,
            ),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(self.PIXEL_SIZE, self.PIXEL_SIZE),
            image_size=(self.RECONST_SIZE, self.RECONST_SIZE),
            trusted_range=(0, 65535),
            # px_mm = px_mm,
            mask=[],
        )

    def _beam(self):
        h5_handle = h5py.File(self.image_filename, "r")
        eV = h5_handle[self.tag]["photon_energy_ev"][()]
        h5_handle.close()

        return self._beam_factory.simple(factor_ev_angstrom / eV)

    def get_num_images(self):
        return len(self._images)

    def get_raw_data(self, index=None):
        if index is not None and self.index != index:
            self.set_index(index)

        if self._raw_data is None:
            h5_handle = h5py.File(self.image_filename, "r")
            data = h5_handle[self.tag]["data"][()].astype(np.int32)
            h5_handle.close()
            self._raw_data = flex.int(data)

        return self._raw_data

    def get_detector(self, index=None):
        if self._detector_instance is None:
            self._detector_instance = self._detector()

        return self._detector_instance

    def get_static_mask(self):
        # This means when the pixel mask is present, trusted region is ignored.
        # The used provided masks (if any) will be automatically merged.
        # see https://github.com/dials/dials/issues/236
        return self.mask

    def get_beam(self, index=None):
        if index is not None and self.index != index:
            self.set_index(index)
            self._beam_instance = None

        if self._beam_instance is None:
            self._beam_instance = self._beam()

        return self._beam_instance


if __name__ == "__main__":
    print(FormatHDF5SaclaRayonix.understand(sys.argv[1]))
    FormatHDF5SaclaRayonix(sys.argv[1])



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for example on ALS beamline 8.2.1 from back in the
day which had its own way of recording beam centre.
"""

from __future__ import annotations

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN445(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 445."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 445."""

        # check this is detector serial number 445

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        return int(header["DETECTOR_SN"]) == 445

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        beam_x = float(self._header_dictionary["DENZO_XBEAM"])
        beam_y = float(self._header_dictionary["DENZO_YBEAM"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )
        trusted_range = self._adsc_trusted_range()

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            trusted_range,
            [],
            gain=self._adsc_module_gain(),
            pedestal=40,
        )



from __future__ import annotations

import copy
import pickle
import sys

from iotbx.detectors.npy import image_dict_to_unicode

from dxtbx.format.FormatPYunspecified import (
    FormatPYunspecified,
    FormatPYunspecifiedInMemory,
)
from dxtbx.format.FormatStill import FormatStill


class FormatPYunspecifiedStill(FormatStill, FormatPYunspecified):
    @staticmethod
    def understand(image_file):
        """Seems like the static method wastes a lot of effort here; it's not possible to
        just read the first few bytes; instead understand() reads the entire first data
        item in the file; an entire binary image.  This data is then read again in the
        _start() method and again in the detectorbase constructor.
        """

        try:
            with FormatPYunspecified.open_file(image_file, "rb") as fh:
                data = image_dict_to_unicode(pickle.load(fh, encoding="bytes"))
        except OSError:
            return False

        if "OSC_START" not in data or "OSC_RANGE" not in data:
            return True

        return data["OSC_RANGE"] <= 0


class FormatPYunspecifiedStillInMemory(FormatStill, FormatPYunspecifiedInMemory):
    """Overrides the Format object's init method to accept an image dictionary
    instead of a file name. Used with XFELs when it is desirable to never write
    a file to disk, but to process it only in memory.
    """

    @staticmethod
    def understand(image_file):
        """If it's an image dictionary, we understand this"""
        data = image_file
        if not isinstance(data, dict):
            return False
        if "OSC_START" not in data or "OSC_RANGE" not in data:
            return True
        return data["OSC_RANGE"] <= 0

    def __init__(self, data, **kwargs):
        """@param data In memory image dictionary, already initialized"""
        super().__init__(data, **kwargs)
        self._image_file = copy.deepcopy(data)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatPYunspecifiedStill.understand(arg))



from __future__ import annotations

import pytest

from dxtbx.format.FormatCBFMiniEigerDLS16MSN160 import FormatCBFMiniEigerDLS16MSN160
from dxtbx.masking import SmarGonShadowMasker
from dxtbx.model.experiment_list import ExperimentListFactory


def test_dlsnxs2cbf_therm(dials_data):
    filename = dials_data("image_examples") / "dlsnxs2cbf_therm_0001.cbf.gz"

    assert FormatCBFMiniEigerDLS16MSN160.understand(filename)
    expts = ExperimentListFactory.from_filenames(
        [filename], format_kwargs={"dynamic_shadowing": True}
    )
    assert len(expts) == 1
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatCBFMiniEigerDLS16MSN160
    assert imageset.has_dynamic_mask()
    gonio = imageset.get_goniometer()
    assert list(gonio.get_angles()) == pytest.approx([0.0, 0.0, 174.0])
    assert list(gonio.get_axes().as_double()) == pytest.approx(
        [1.0, 0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0]
    )
    assert list(gonio.get_names()) == ["GON_PHI", "GON_CHI", "GON_OMEGA"]
    masker = imageset.masker()
    assert isinstance(masker, SmarGonShadowMasker)
    assert masker.get_mask(imageset.get_detector(), 0)[0].count(False) == 0



from __future__ import annotations

from libtbx.test_utils import approx_equal
from rstbx.cftbx import coordinate_frame_helpers

from dxtbx.imageset import ImageSetFactory
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.serialize import xds


def test_to_xds(dials_data, tmpdir):
    file_names = dials_data("centroid_test_data").glob("centroid_00*.cbf")
    sequence = ImageSetFactory.new(list(file_names))[0]
    to_xds = xds.to_xds(sequence)
    s1 = to_xds.XDS_INP()
    expected_f = """\
DETECTOR=PILATUS MINIMUM_VALID_PIXEL_VALUE=0 OVERLOAD=495976
SENSOR_THICKNESS= 0.320
!SENSOR_MATERIAL / THICKNESS Si 0.320
!SILICON= %%s
DIRECTION_OF_DETECTOR_X-AXIS= 1.00000 0.00000 0.00000
DIRECTION_OF_DETECTOR_Y-AXIS= 0.00000 1.00000 0.00000
NX=2463 NY=2527 QX=0.1720 QY=0.1720
DETECTOR_DISTANCE= 190.180000
ORGX= 1235.84 ORGY= 1279.58
ROTATION_AXIS= 1.00000 0.00000 0.00000
STARTING_ANGLE= 0.000
OSCILLATION_RANGE= 0.200
X-RAY_WAVELENGTH= 0.97950
INCIDENT_BEAM_DIRECTION= -0.000 -0.000 1.021
FRACTION_OF_POLARIZATION= 0.999
POLARIZATION_PLANE_NORMAL= 0.000 1.000 0.000
NAME_TEMPLATE_OF_DATA_FRAMES= %s
TRUSTED_REGION= 0.0 1.41
UNTRUSTED_RECTANGLE= 487 495 0 2528
UNTRUSTED_RECTANGLE= 981 989 0 2528
UNTRUSTED_RECTANGLE= 1475 1483 0 2528
UNTRUSTED_RECTANGLE= 1969 1977 0 2528
UNTRUSTED_RECTANGLE= 0 2464 195 213
UNTRUSTED_RECTANGLE= 0 2464 407 425
UNTRUSTED_RECTANGLE= 0 2464 619 637
UNTRUSTED_RECTANGLE= 0 2464 831 849
UNTRUSTED_RECTANGLE= 0 2464 1043 1061
UNTRUSTED_RECTANGLE= 0 2464 1255 1273
UNTRUSTED_RECTANGLE= 0 2464 1467 1485
UNTRUSTED_RECTANGLE= 0 2464 1679 1697
UNTRUSTED_RECTANGLE= 0 2464 1891 1909
UNTRUSTED_RECTANGLE= 0 2464 2103 2121
UNTRUSTED_RECTANGLE= 0 2464 2315 2333
DATA_RANGE= 1 9
JOB=XYCORR INIT COLSPOT IDXREF DEFPIX INTEGRATE CORRECT\
""" % (dials_data("centroid_test_data") / "centroid_????.cbf")

    # universe changed once, so be flexible
    expected = [expected_f % a for a in ["3.960382", "3.960386"]]

    assert s1 in expected
    real_space_a = (-5.327642, -39.034747, -4.988286)
    real_space_b = (-35.253495, 7.596265, -22.127661)
    real_space_c = (-22.673623, -1.486119, 35.793463)
    s2 = to_xds.xparm_xds(real_space_a, real_space_b, real_space_c, space_group=1)
    # run coordinate frame converter on xparm.xds as a sanity check
    with open(tmpdir / "xparm.xds", mode="wb") as fh:
        fh.write(s2.encode("ASCII"))

    converter = coordinate_frame_helpers.import_xds_xparm(tmpdir / "xparm.xds")
    detector = sequence.get_detector()
    goniometer = sequence.get_goniometer()
    beam = sequence.get_beam()
    assert approx_equal(real_space_a, converter.get_real_space_a())
    assert approx_equal(real_space_b, converter.get_real_space_b())
    assert approx_equal(real_space_c, converter.get_real_space_c())
    assert approx_equal(goniometer.get_rotation_axis(), converter.get_rotation_axis())
    assert approx_equal(
        beam.get_sample_to_source_direction(), converter.get_sample_to_source().elems
    )
    assert approx_equal(detector[0].get_fast_axis(), converter.get_detector_fast())
    assert approx_equal(detector[0].get_slow_axis(), converter.get_detector_slow())
    assert approx_equal(detector[0].get_origin(), converter.get_detector_origin())


def test_to_xds_multi_panel_i23(dials_data, tmpdir, mocker):
    data_dir = dials_data("image_examples")
    file_name = str(data_dir / "DLS_I23-germ_13KeV_0001.cbf.bz2")
    sequence = ImageSetFactory.new([file_name])[0]
    to_xds = xds.to_xds(sequence)
    s1 = to_xds.XDS_INP()
    for expected_substr in (
        """\
!
! SEGMENT 1
!
SEGMENT= 1 2463 1 195
DIRECTION_OF_SEGMENT_X-AXIS= 1.00000 0.00000 0.00000
DIRECTION_OF_SEGMENT_Y-AXIS= 0.00000 -0.14347 0.98966
SEGMENT_DISTANCE= 250.000
SEGMENT_ORGX= 1075.00 SEGMENT_ORGY= 97.67""",
        """\
!
! SEGMENT 24
!
SEGMENT= 1 2463 4877 5071
DIRECTION_OF_SEGMENT_X-AXIS= 1.00000 0.00000 0.00000
DIRECTION_OF_SEGMENT_Y-AXIS= 0.00000 -0.06390 -0.99796
SEGMENT_DISTANCE= 250.000
SEGMENT_ORGX= 1075.00 SEGMENT_ORGY= 4973.67""",
    ):
        assert expected_substr in s1
    assert "UNTRUSTED_RECTANGLE" not in s1

    # Fool the format class into masking out a couple of bad modules
    mocked_timestamp = mocker.patch(
        "dxtbx.format.FormatCBFMiniPilatusDLS12M.get_pilatus_timestamp"
    )
    mocked_timestamp.return_value = 1574857526.34
    sequence = ImageSetFactory.new([file_name])[0]
    to_xds = xds.to_xds(sequence)
    s = to_xds.XDS_INP()
    assert "UNTRUSTED_RECTANGLE= 0 488 3604 3800" in s
    assert "UNTRUSTED_RECTANGLE= 1976 2464 3604 3800" in s


def test_vmxi_thaumatin(dials_data):
    master_h5 = dials_data("vmxi_thaumatin") / "image_15799_master.h5"
    expts = ExperimentListFactory.from_filenames([master_h5])
    to_xds = xds.to_xds(expts[0].imageset)
    s = to_xds.XDS_INP()
    assert "DETECTOR=EIGER" in s
    assert "SENSOR_THICKNESS= 0.450" in s


def test_xds_to_imageset(tmp_path, dials_data):
    input_file = """
DATA_RANGE=1 100
DETECTOR=EIGER
DETECTOR_DISTANCE=259.000000
DIRECTION_OF_DETECTOR_X-AXIS=1.00000 0.00000 0.00000
DIRECTION_OF_DETECTOR_Y-AXIS=0.00000 1.00000 0.00000
FRACTION_OF_POLARIZATION=0.999
FRIEDEL'S_LAW=TRUE
INCIDENT_BEAM_DIRECTION=-0.000 -0.000 1.049
MINIMUM_VALID_PIXEL_VALUE=0
NAME_TEMPLATE_OF_DATA_FRAMES=thaumatin_??????.h5
NX=4148
NY=4362
ORGX=2050.50
ORGY=2150.50
OSCILLATION_RANGE=0.100
OVERLOAD=126952
POLARIZATION_PLANE_NORMAL=0.000 1.000 0.000
QX=0.0750
QY=0.0750
ROTATION_AXIS=0.00000 -1.00000 -0.00000
STARTING_ANGLE=200.000
TRUSTED_REGION=0.0 1.41
SENSOR_THICKNESS=0.450
X-RAY_WAVELENGTH=0.95372
SPACE_GROUP_NUMBER=89
UNIT_CELL_CONSTANTS=57.500000 57.500000 149.000000 90.000000 90.000000 90.000000
"""
    xds_inp = tmp_path / "XDS.INP"
    with open(xds_inp, "w") as f:
        f.write(input_file)
    ## Note that the XDS.INP is for thaumatin, and centroid test data is not, but
    ## this is sufficient for testing the creation of an imageset of the correct size.
    xds_other = dials_data("centroid_test_data") / "INTEGRATE.HKL"
    iset = xds.to_imageset(xds_inp, xds_other)
    assert len(iset) == 100



# LIBTBX_SET_DISPATCHER_NAME dxtbx.radial_average
# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1

from __future__ import annotations

import math
import os
import sys

import numpy as np

import cctbx.miller
import iotbx.phil
from cctbx.crystal import symmetry
from libtbx import easy_pickle
from libtbx.utils import Sorry, Usage
from scitbx.array_family import flex
from scitbx.matrix import col

import dxtbx
import dxtbx.util
from dxtbx.ext import radial_average
from dxtbx.model.experiment_list import ExperimentListFactory

master_phil = iotbx.phil.parse(
    """
  file_path = None
    .type = str
    .multiple = True
    .help = Files to read
  n_bins = 0
    .type = int
    .help = Number of bins in the radial average. Auto determined if set to 0
  verbose = True
    .type = bool
    .help = Extra logging information
  output_bins = True
    .type = bool
    .help = Whether to print values for each bin
  output_file = None
    .type = str
    .help = Output file for logging results
  plot_x_max = None
    .type = int
    .help = Max value for x axis
  plot_y_max = None
    .type = int
    .help = Max value for xyaxis
  low_max_two_theta_limit = None
    .type = float
    .help = Low two theta cutoff
  normalize = False
    .type = bool
    .help = Whether to normalize the Y values to 1
  show_plots = True
    .type = bool
    .help = Whether to show the radial average plot
  mask = None
    .type = str
    .help = DIALS style pixel mask. Average will skip these pixels
  median_filter_size = None
    .type = int
    .help = If not none, applies a scipy.ndimage median_filter to the average
  x_axis = *two_theta q resolution
    .type = choice
    .help = Units for x axis
  image_number = None
    .type = int
    .help = When supplying a composite file, which image to show. Otherwise \
            shows all images in the file.
  panel = None
    .type = int
    .help = Only use data from the specified panel
  max_images = None
    .type = int
    .help = When supplying a composite file, only show up to max_images images
  reference_geometry = None
    .type = path
    .help = Apply this geometry before creating average
  unit_cell = None
    .type = unit_cell
    .help = Show positions of miller indices from this unit_cell and space \
            group.
  space_group = None
    .type = space_group
    .help = Show positions of miller indices from this unit_cell and space \
            group.
  d_min = 4
    .type = float
    .help = Resolution limit for miller index display
"""
)


def run(args=None, imageset=None):
    dxtbx.util.encode_output_as_utf8()
    args = sys.argv[1:] if args is None else args

    # Parse input
    try:
        len(args)
    except TypeError:
        params = args
    else:
        user_phil = []
        for arg in args:
            if "=" in arg:
                try:
                    user_phil.append(iotbx.phil.parse(arg))
                except RuntimeError as e:
                    raise Sorry(
                        "Unrecognized argument '%s' (error: %s)" % (arg, str(e))
                    )
            else:
                try:
                    user_phil.append(iotbx.phil.parse("""file_path=%s""" % arg))
                except ValueError:
                    raise Sorry("Unrecognized argument '%s'" % arg)
        params = master_phil.fetch(sources=user_phil).extract()
    if imageset is None:
        if (
            params.file_path is None
            or len(params.file_path) == 0
            or not all(os.path.isfile(f) for f in params.file_path)
        ):
            master_phil.show(attributes_level=2)
            raise Usage(
                "file_path must be defined (either file_path=XXX, or the path alone)."
            )
    assert params.n_bins is not None
    assert params.verbose is not None
    assert params.output_bins is not None

    if params.reference_geometry is not None:
        ref_expts = ExperimentListFactory.from_json_file(
            params.reference_geometry, check_format=None
        )
        assert len(ref_expts.detectors()) == 1, (
            "Provide only one detector in the reference geometry file"
        )
        detector = ref_expts.detectors()[0]

    # Allow writing to a file instead of stdout
    if params.output_file is None:
        logger = sys.stdout
    else:
        logger = open(params.output_file, "w")
        logger.write("%s " % params.output_file)

    if params.show_plots:
        from matplotlib import pyplot as plt

        colormap = plt.cm.gist_ncar
        plt.gca().set_prop_cycle(
            color=[colormap(i) for i in np.linspace(0, 0.9, len(params.file_path))]
        )

    if params.mask is not None and not isinstance(params.mask, tuple):
        params.mask = easy_pickle.load(params.mask)

    if imageset is None:
        iterable = params.file_path

        def load_func(x):
            try:
                obj = dxtbx.load(x).get_imageset([x])
            except (TypeError, ValueError):
                obj = ExperimentListFactory.from_json_file(x)[0].imageset
            return obj

    else:
        iterable = [imageset]

        def load_func(x):
            return x

    # Iterate over each file provided
    for item in iterable:
        iset = load_func(item)
        n_images = len(iset)
        if params.image_number is None:
            if params.max_images is None:
                subiterable = range(n_images)
            else:
                subiterable = range(0, min(params.max_images, n_images))
        else:
            subiterable = [params.image_number]
        for image_number in subiterable:
            try:
                beam = iset.get_beam(image_number)
            except Exception:
                print("Beam not found for image %d" % image_number)
                continue
            if params.reference_geometry is None:
                detector = iset.get_detector(image_number)
            s0 = col(beam.get_s0())

            # Search the detector for the panel farthest from the beam. The
            # number of bins in the radial average will be equal to the
            # farthest point from the beam on the detector, in pixels, unless
            # overridden at the command line
            panel_res = [p.get_max_resolution_at_corners(s0) for p in detector]
            farthest_panel = detector[panel_res.index(min(panel_res))]
            size2, size1 = farthest_panel.get_image_size()
            corners = [(0, 0), (size1 - 1, 0), (0, size2 - 1), (size1 - 1, size2 - 1)]
            corners_lab = [col(farthest_panel.get_pixel_lab_coord(c)) for c in corners]
            corner_two_thetas = [
                farthest_panel.get_two_theta_at_pixel(s0, c) for c in corners
            ]
            extent_two_theta = max(corner_two_thetas)
            max_corner = corners_lab[corner_two_thetas.index(extent_two_theta)]
            extent = int(
                math.ceil(
                    max_corner.length()
                    * math.sin(extent_two_theta)
                    / max(farthest_panel.get_pixel_size())
                )
            )
            extent_two_theta *= 180 / math.pi

            if params.n_bins < extent:
                params.n_bins = extent

            # These arrays will store the radial average info
            sums = flex.double(params.n_bins) * 0
            sums_sq = flex.double(params.n_bins) * 0
            counts = flex.int(params.n_bins) * 0

            all_data = iset[image_number]

            if not isinstance(all_data, tuple):
                all_data = (all_data,)

            for tile, (panel, data) in enumerate(zip(detector, all_data)):
                if params.panel is not None and tile != params.panel:
                    continue

                if params.mask is None:
                    mask = flex.bool(flex.grid(data.focus()), True)
                else:
                    mask = params.mask[tile]

                if hasattr(data, "as_double"):
                    data = data.as_double()

                logger.flush()
                if params.verbose:
                    logger.write(
                        "Average intensity tile %d: %9.3f\n" % (tile, flex.mean(data))
                    )
                    logger.write("N bins: %d\n" % params.n_bins)
                    logger.flush()

                x1, y1, x2, y2 = (
                    0,
                    0,
                    panel.get_image_size()[1],
                    panel.get_image_size()[0],
                )
                bc = panel.get_beam_centre_px(beam.get_s0())
                bc = int(round(bc[1])), int(round(bc[0]))

                # compute the average
                radial_average(
                    data,
                    mask,
                    bc,
                    sums,
                    sums_sq,
                    counts,
                    panel.get_pixel_size()[0],
                    panel.get_distance(),
                    (x1, y1),
                    (x2, y2),
                )

            # average the results, avoiding division by zero
            results = sums.set_selected(counts <= 0, 0)
            results /= counts.set_selected(counts <= 0, 1).as_double()

            if params.median_filter_size is not None:
                logger.write(
                    "WARNING, the median filter is not fully propagated to the variances\n"
                )
                from scipy.ndimage.filters import median_filter

                results = flex.double(
                    median_filter(
                        results.as_numpy_array(), size=params.median_filter_size
                    )
                )

            # calculate standard devations
            stddev_sel = ((sums_sq - sums * results) >= 0) & (counts > 0)
            std_devs = flex.double(len(sums), 0)
            std_devs.set_selected(
                stddev_sel,
                (
                    sums_sq.select(stddev_sel)
                    - sums.select(stddev_sel) * results.select(stddev_sel)
                )
                / counts.select(stddev_sel).as_double(),
            )
            std_devs = flex.sqrt(std_devs)

            twotheta = (
                flex.double(range(len(results))) * extent_two_theta / params.n_bins
            )
            q_vals = (
                4 * math.pi * flex.sin(math.pi * twotheta / 360) / beam.get_wavelength()
            )
            # nlmbda = 2dsin(theta)
            resolution = flex.double(len(twotheta), 0)
            nonzero = twotheta > 0
            resolution.set_selected(
                nonzero,
                beam.get_wavelength()
                / (2 * flex.asin((math.pi / 180) * twotheta.select(nonzero) / 2)),
            )

            if params.low_max_two_theta_limit is None:
                subset = results
            else:
                subset = results.select(twotheta >= params.low_max_two_theta_limit)

            max_result = flex.max(subset)

            if params.x_axis == "two_theta":
                xvals = twotheta
                max_x = twotheta[flex.first_index(results, max_result)]
            elif params.x_axis == "q":
                xvals = q_vals
                max_x = q_vals[flex.first_index(results, max_result)]
            elif params.x_axis == "resolution":
                xvals = resolution
                max_x = resolution[flex.first_index(results, max_result)]

            for i, r in enumerate(results):
                val = xvals[i]
                if params.output_bins and "%.3f" % r != "nan":
                    # logger.write("%9.3f %9.3f\n"%     (val,r))        #.xy  format for Rex.cell.
                    logger.write(
                        "%9.3f %9.3f %9.3f\n" % (val, r, std_devs[i])
                    )  # .xye format for GSASII
                # logger.write("%.3f %.3f %.3f\n"%(val,r,ds[i]))  # include calculated d spacings
            logger.write(f"Maximum {params.x_axis}: {max_x:f}, value: {max_result:f}\n")

            if params.show_plots:
                if params.plot_x_max is not None:
                    if params.x_axis == "resolution":
                        results = results.select(xvals >= params.plot_x_max)
                        xvals = xvals.select(xvals >= params.plot_x_max)
                    else:
                        results = results.select(xvals <= params.plot_x_max)
                        xvals = xvals.select(xvals <= params.plot_x_max)
                if params.x_axis == "resolution":
                    xvals = 1 / (xvals**2)
                if params.normalize:
                    plt.plot(
                        xvals.as_numpy_array(),
                        (results / flex.max(results)).as_numpy_array(),
                        "-",
                    )
                else:
                    plt.plot(xvals.as_numpy_array(), results.as_numpy_array(), "-")
                if params.x_axis == "two_theta":
                    plt.xlabel("2 theta")
                elif params.x_axis == "q":
                    plt.xlabel("q")
                elif params.x_axis == "resolution":
                    from matplotlib.ticker import FuncFormatter

                    plt.xlabel("Resolution ($\\AA$)")

                    def resolution(x, pos):
                        if x <= 0:
                            return "-"
                        return "%.1f" % (1 / math.sqrt(x))

                    formatter = FuncFormatter(resolution)
                    plt.gca().xaxis.set_major_formatter(formatter)
                plt.ylabel("Avg ADUs")
                if params.plot_y_max is not None:
                    plt.ylim(0, params.plot_y_max)

    if params.show_plots:
        if params.unit_cell or params.space_group:
            assert params.unit_cell and params.space_group
            sym = symmetry(
                unit_cell=params.unit_cell, space_group=params.space_group.group()
            )
            hkl_list = cctbx.miller.build_set(sym, False, d_min=params.d_min)
            tt = sym.unit_cell().two_theta(
                hkl_list.indices(), beam.get_wavelength(), deg=True
            )
            if params.x_axis == "q":
                q_vals = (
                    4 * math.pi * flex.sin(math.pi * tt / 360) / beam.get_wavelength()
                )
                vals = q_vals
            elif params.x_axis == "resolution":
                # nlmbda = 2dsin(theta)
                resolution = flex.double(len(tt), 0)
                nonzero = tt > 0
                resolution.set_selected(
                    nonzero,
                    beam.get_wavelength()
                    / (2 * flex.asin((math.pi / 180) * tt.select(nonzero) / 2)),
                )
                vals = resolution
                vals = 1 / (vals**2)
            elif params.x_axis == "two_theta":
                vals = tt

            y_max = 1 if params.normalize else 50
            for val in vals:
                plt.plot([val, val], [0, y_max], "r")

        # plt.legend([os.path.basename(os.path.splitext(f)[0]) for f in params.file_path], ncol=2)
        plt.show()

    return xvals, results


if __name__ == "__main__":
    run()



"""Set up for Soleil PX1, with full kappa goniometer"""

from __future__ import annotations

import sys

from scitbx.array_family import flex

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus


class FormatCBFMiniPilatusSOLEILPX16MSN106(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus images for 6M SN 106 @
    Synchrotrol Soleil PX1."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0106, Soleil" in record
            ):
                return True

        return False

    def _goniometer(self):
        """Construct a goniometer from the records in the mini CBF header."""

        if (
            "Alpha" in self._cif_header_dictionary
            and "Kappa" in self._cif_header_dictionary
        ):
            # Kappa
            alpha = float(self._cif_header_dictionary["Alpha"].split()[0])
            omega = float(self._cif_header_dictionary["Chi"].split()[0])
            kappa = float(self._cif_header_dictionary["Kappa"].split()[0])
            phi = float(self._cif_header_dictionary["Phi"].split()[0])

            axis = self._cif_header_dictionary["Oscillation_axis"]

            scanaxis = {"OMEGA": "Omega", "PHI": "Phi"}

            assert axis in scanaxis

            # this is the direction the arm points in at datum
            direction = "+z"

            return self._goniometer_factory.make_kappa_goniometer(
                alpha, omega, kappa, phi, direction, scanaxis[axis]
            )

        else:
            # Smargon
            phi = float(self._cif_header_dictionary["Phi"].split()[0])
            chi = float(self._cif_header_dictionary["Chi"].split()[0])
            omega = float(self._cif_header_dictionary["Omega"].split()[0])

            names = flex.std_string(("PHI", "CHI", "OMEGA"))
            axes = flex.vec3_double(((1, 0, 0), (0, 0, -1), (1, 0, 0)))
            angles = flex.double((phi, chi, omega))

            axis = self._cif_header_dictionary["Oscillation_axis"].upper()
            assert axis in names, axis
            scan_axis = flex.first_index(names, axis)

            return self._goniometer_factory.make_multi_axis_goniometer(
                axes, angles, names, scan_axis
            )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusSOLEILPX16MSN106.understand(arg))



from __future__ import annotations

import bz2

import dxtbx


def test_noninteger_pedestal(dials_data, tmpdir):
    data_dir = dials_data("image_examples")
    filename = str(data_dir / "APS_14BMC-q315_1_001.img.bz2")
    # Read this file in as data
    with bz2.open(filename, "rb") as f:
        data = f.read()

    # Write out with an inserted header item of a noninteger pedestal
    test_file = tmpdir / "test_pedestal_001.img"
    with test_file.open("wb") as f:
        f.write(data.replace(b"DIM=2;\n", b"DIM=2;\nIMAGE_PEDESTAL=42.0;\n"))

    # Make sure this loads
    dxtbx.load(test_file)



from __future__ import annotations

import iotbx.cif.model

import dxtbx.data.beamline_defs


class get_definition(dxtbx.data.beamline_defs.template):
    def __init__(self, timestamp=None, **kwargs):
        self._timestamp = timestamp

    def CIF_block(self):
        """Interface function to generate a CIF block for this detector."""
        return self._identify_time()(mmcif=False)

    def mmCIF_block(self):
        """Interface function to generate an mmCIF block for this detector."""
        return self._identify_time()(mmcif=True)

    def _identify_time(self):
        """Determine detector environment based on timestamp."""
        if not self._timestamp:  # default to I19
            return self._at_I19

        if self._timestamp >= self._date_to_epoch(2015, 11, 1):
            return self._at_I19  # moved to I19 on 01.11.2015

        return self._at_I04_1  # before was on I04-1

    def _base(self, mmcif=False):
        """Generates
        1. a CIF/mmCIF block that contains information that
        is always true about the detector.
        2. a lookup function for CIF/mmCIF strings."""
        # prepare string lookup table
        lookup = self._lookup(mmcif)

        b = iotbx.cif.model.block()
        b[lookup("df.detector")] = "Photon counting pixel array"
        b[lookup("df.detector.type")] = "Dectris PILATUS 2M"
        b[lookup("df.src")] = "Synchrotron"

        return b, lookup

    def _at_I19(self, mmcif=False):
        b, lookup = self._base(mmcif)

        b[lookup("df.m.dev")] = "Fixed \\c 3-circle diffractometer"
        b[lookup("df.m.dev.type")] = "Fluid Film Devices"
        b[lookup("df.m.method")] = "shutterless scans"
        b[lookup("df.m.spec.supp")] = "MiTeGen MicroMount"
        b[lookup("df.rad.mono")] = "Silicon 111"
        b[lookup("df.src.details")] = "Allan et al. (2017)"
        b[lookup("df.src.type")] = "Diamond Light Source Beamline I19-1"
        b[lookup("references")] = "Allan, D. R. et al. (2017) Crystals, 7(11), 336."
        b[lookup("sw.collection")] = "GDA - generic data acquisition software"

        return b

    def _at_I04_1(self, mmcif=False):
        b, lookup = self._base(mmcif)

        b[lookup("df.src.type")] = "Diamond Light Source Beamline I04-1"

        return b



from __future__ import annotations

import sys

from dxtbx.format.FormatPYunspecified import FormatPYunspecified


class FormatPYCXI32(FormatPYunspecified):
    """PREFERENCE FILE.
    Treats any Pickle-format file lacking a DETECTOR_FORMAT_VERSION key
    automatically as format CXI 3.2.
    Rename this file to FormatPYCXI32.py and put in dxtbx/format path.
    """

    def _start(self):
        FormatPYunspecified.start_helper(
            self, version_token="distl.detector_format_version=CXI 3.2"
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatPYCXI32.understand(arg))



from __future__ import annotations

import calendar
import datetime
import struct

from dxtbx.format.Format import Format
from dxtbx.format.FormatRAXIS import RAXISHelper


class FormatRAXISIVSPring8(RAXISHelper, Format):
    """Format class for R-AXIS4 images. Currently the only example we have is
    from SPring-8, which requires a reverse axis goniometer. It is not clear how
    to distinguish this detector from others that produce 'R-AXIS4' images that we
    may come across in future. This will be a problem if other detectors do not
    also require reverse axis.
    """

    @staticmethod
    def understand(image_file):
        try:
            with Format.open_file(image_file, "rb") as fh:
                header = fh.read(1024)
        except OSError:
            return False

        # A few items expected to be the same from image to image that we can use
        # as a fingerprint for this instrument
        if header[0:7] != b"R-AXIS4":
            return False

        # We expect an invalid goniometer section, indicated by wrong magic number
        if struct.unpack(">i", header[852:856]) == 1:
            return False

        if header[812:822].strip() != b"IRIS":
            return False

        return True

    def _goniometer(self):
        return self._goniometer_factory.single_axis_reverse()

    def _detector(self):
        """Return a model for the detector as defined in the image header,
        with the additional knowledge about how things are arranged i.e. that
        the principle rotation axis vector points from the sample downwards."""

        values = self._detector_helper()

        return self._detector_factory.simple(
            "IMAGE_PLATE",
            values["distance"],
            values["beam"],
            "+x",
            "+y",
            (values["dx"], values["dy"]),
            (values["nx"], values["ny"]),
            (0, 1000000),
            [],
        )

    def _beam(self):
        """Return a simple model for the beam."""

        wavelength = struct.unpack(self._f, self._header_bytes[292:296])[0]

        return self._beam_factory.simple(wavelength)

    def _scan(self):
        """Return the scan information for this image."""
        f = self._f
        header = self._header_bytes

        exposure_time = struct.unpack(f, header[536:540])[0]

        y, m, d = map(int, header[256:268].strip().split(b"-"))

        epoch = calendar.timegm(datetime.datetime(y, m, d, 0, 0, 0).timetuple())

        # For this instrument, the header goniometer section is invalid. 3 floats
        # starting at byte 520 specify phi0, phis and phie.
        # (http://www.rigaku.com/downloads/software/readimage.html)

        osc_dat = struct.unpack(f, header[520:524])[0]
        osc_start = osc_dat + struct.unpack(f, header[524:528])[0]
        osc_end = osc_dat + struct.unpack(f, header[528:532])[0]

        osc_range = osc_end - osc_start

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )



from __future__ import annotations

from dxtbx.command_line import print_header


def test_print_header(dials_data, capsys):
    screen = dials_data("thaumatin_eiger_screen")
    master = screen / "Therm_6_1_master.h5"
    print_header.run([str(master)])

    expected_output = [
        f"=== {master} ===",
        "Using header reader: FormatNXmxDLS16M",
    ]

    captured = capsys.readouterr()
    assert not captured.err
    for record in expected_output:
        assert record.strip() in captured.out, record



from __future__ import annotations

from typing import Any

from scitbx.array_family import flex

def dataset_as_flex_double(dataset_id: Any, selection: tuple[slice]) -> flex.double: ...
def dataset_as_flex_float(dataset_id: Any, selection: tuple[slice]) -> flex.float: ...
def dataset_as_flex_int(dataset_id: Any, selection: tuple[slice]) -> flex.int: ...



from __future__ import annotations

import logging

import h5py
import numpy as np

import cctbx.array_family.flex as flex

import dxtbx_flumpy as flumpy
from dxtbx import IncorrectFormatError
from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.model import Detector
from dxtbx.model.beam import BeamFactory, PolychromaticBeam, Probe
from dxtbx.model.goniometer import Goniometer, GoniometerFactory
from dxtbx.model.scan import Scan, ScanFactory
from dxtbx.model.tof_helpers import wavelength_from_tof

logger = logging.getLogger(__name__)


class FormatESSNMX(FormatHDF5):
    """
    Class to read files from NMX
    https://europeanspallationsource.se/instruments/nmx
    preprocessed files in scipp to obtain binned data
    """

    def __init__(self, image_file, **kwargs) -> None:
        if not FormatESSNMX.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self._nxs_file = h5py.File(image_file, "r")
        self._raw_data = None

    @staticmethod
    def understand(image_file: str) -> bool:
        try:
            return FormatESSNMX.is_nmx_file(image_file)
        except (OSError, KeyError):
            return False

    @staticmethod
    def is_nmx_file(image_file: str) -> bool:
        def get_name(image_file):
            try:
                with h5py.File(image_file, "r") as handle:
                    return handle["entry/instrument/name"][...].item().decode()
            except (OSError, KeyError, AttributeError):
                return ""

        def is_McStas(image_file: str) -> bool:
            try:
                with h5py.File(image_file, "r") as handle:
                    if handle.get(
                        "entry/metadata/mcstas_weight2count_scale_factor", None
                    ):
                        return True
                    return False
            except (OSError, KeyError, AttributeError):
                return False

        return get_name(image_file) == "NMX" and not is_McStas(image_file)

    def get_instrument_name(self) -> str:
        return "NMX"

    def get_experiment_description(self) -> str:
        return "NMX Data"

    def _load_raw_data(self) -> None:
        raw_data = []
        for panel in self._get_panels():
            spectra = panel["data"][...]
            raw_data.append(flumpy.from_numpy(np.ascontiguousarray(spectra)))

        self._raw_data = tuple(raw_data)

    def get_raw_data(
        self, index: int, use_loaded_data: bool = False
    ) -> tuple[flex.int]:
        raw_data = []
        # image_size = self._get_image_size()
        # total_pixels = image_size[0] * image_size[1]

        if use_loaded_data:
            if self._raw_data is None:
                self._load_raw_data()
            for panel in self._raw_data:
                data = panel[:, :, index : index + 1].astype(np.int32)
                data.reshape(flex.grid(panel.all()[0], panel.all()[1]))
                data.matrix_transpose_in_place()
                raw_data.append(data)

        else:
            for panel in self._get_panels():
                spectra = panel["data"][:, :, index].astype(np.int32)
                raw_data.append(flumpy.from_numpy(np.ascontiguousarray(spectra)))

        return tuple(raw_data)

    def _get_time_channel_bins(self) -> list[float]:
        # (usec)
        # the tofs are recorded separately per panel but they
        # should all be the same
        for panel in self._get_panels():
            return panel["time_of_flight"][...] / 1e3

    def _get_time_of_flight(self) -> list[float]:
        # (usec)
        bins = self._get_time_channel_bins()
        return [float((bins[i] + bins[i + 1]) * 0.5) for i in range(len(bins) - 1)]

    def get_num_images(self) -> int:
        return len(self._get_time_of_flight())

    def get_detector(self, index: int = None) -> Detector:
        panel_names = self._get_panel_names()
        panel_type = self._get_panel_type()
        trusted_range = self._get_panel_trusted_range()
        pixel_size = self._get_pixel_size()
        gain = self._get_panel_gain()
        detector = Detector()
        root = detector.hierarchy()
        panels = self._get_panels()
        panel_projections = self._get_panel_projections_2d(panels)
        for panel_dset, panel_name in zip(panels, panel_names):
            panel = root.add_panel()
            panel.set_type(panel_type)
            panel.set_name(panel_name)
            panel.set_image_size(
                self._get_image_size(panel_dset)
            )  # XXX fix to include detectors possibly not being the same size
            panel.set_trusted_range(trusted_range)
            panel.set_pixel_size(pixel_size)
            fast_axis = self._get_panel_fast_axes(panel_dset)
            slow_axis = self._get_panel_slow_axes(panel_dset)
            panel_origin = self._get_panel_origins(panel_dset)
            panel.set_local_frame(fast_axis, slow_axis, panel_origin)

            panel.set_gain(gain)
            i = int(panel_name[-1])
            r, t = panel_projections[i]
            r = tuple(map(int, r))
            t = tuple(map(int, t))
            panel.set_projection_2d(r, t)

        return detector

    def _get_num_panels(self) -> int:
        return len(self._get_panels())

    def _get_panels(self) -> list[h5py._hl.group.Group]:
        """get the detector panel locations in file"""
        panels = []
        inst_dset = self._nxs_file["/entry/instrument/"]
        for _, dset in inst_dset.items():
            if dset.attrs.get("NX_class") == "NXdetector":
                panels.append(dset)
        return panels

    def _get_panel_names(self) -> list[str]:
        panel_names = []
        inst_dset = self._nxs_file["/entry/instrument/"]
        for name, dset in inst_dset.items():
            if dset.attrs.get("NX_class") == "NXdetector":
                panel_names.append(name)
        return panel_names

    def _get_panel_name(self, panel) -> str:
        return panel.name.split("/")[-1]

    def _get_panel_type(self) -> str:
        return "Triple_GEM_Gd"

    def _get_image_size(self, panel) -> tuple[int, int]:
        # (px)
        dset = panel["data"]
        return dset[:, :, 0].shape

    def _get_panel_trusted_range(self) -> tuple[int, int]:
        # 4 * 1280**2 plus buffer
        return (-1, 2**64 - 1)

    def _get_pixel_size(self) -> tuple[float, float]:
        # (mm)
        return (0.4, 0.4)

    def _get_panel_fast_axes(self, panel) -> tuple[float, float, float]:
        # return ((1.0, 0.0, 0.0), (0.0, 0.0, 1.0), (0.0, 0.0, -1.0))
        fast_axis = panel["fast_axis"][...]
        return tuple(fast_axis)

    def _get_panel_slow_axes(self, panel) -> tuple[float, float, float]:
        # return ((0.0, 1.0, 0.0), (0.0, 1.0, 0.0), (0.0, 1.0, 0.0))
        slow_axis = panel["slow_axis"][...]
        return tuple(slow_axis)

    def _get_panel_origins(self, panel) -> tuple[float, float, float]:
        # (mm)
        # return ((-250, -250.0, -292.0), (290, -250.0, -250), (-290, -250.0, 250.0))
        if panel["origin"].attrs.get("first_pixel_position", None) is not None:
            logger.debug("first_pixel_origin found in attrubiute, using...")
            first_pixel_position = panel["origin"].attrs.get("first_pixel_position")
            first_pixel_position *= 1000  # mm
            return tuple(first_pixel_position)
        else:
            origin = panel["origin"][...]
            panelnum = int(panel.name[-1])
            origin *= 1000  # convert to mm
            corrfact = self._get_correction_factor()[panelnum]

            corrorg = origin + corrfact
            return tuple(corrorg)

    def _get_correction_factor(self):
        return np.array([[0, -256, -256], [-256.0, -256.0, 0], [0, -256, 256]])

    def _get_panel_projections_2d(self, panels) -> dict[int : tuple[tuple, tuple]]:
        p_w, p_h = self._get_image_size(panels[0])  # XXX fix later
        p_w += 10
        p_h += 10
        panel_pos = {
            int(panels[0].name[-1]): ((-1, 0, 0, -1), (p_h, 0)),
            int(panels[1].name[-1]): ((-1, 0, 0, -1), (p_h, p_w)),
            int(panels[2].name[-1]): ((-1, 0, 0, -1), (p_h, -p_w)),
        }

        return panel_pos

    def get_beam(self, index: int = None) -> PolychromaticBeam:
        direction = self._get_sample_to_source_direction()
        distance = self._get_sample_to_source_distance()
        wavelength_range = self._get_wavelength_range()
        return BeamFactory.make_polychromatic_beam(
            direction=direction,
            sample_to_source_distance=distance,
            probe=Probe.neutron,
            wavelength_range=wavelength_range,
        )

    def _get_sample_to_source_direction(self) -> tuple[float, float, float]:
        return (0, 0, -1)

    # def _get_wavelength_range(self) -> tuple[float, float]:
    #     # (A)
    #     return (1.8, 2.55)

    def _get_wavelength_range(self) -> tuple[float, float]:
        # (A)
        tofs = np.array(self._get_time_of_flight()) / 1e6  # (s)
        tof_low = tofs[0]
        tof_high = tofs[-1]
        distance = self._get_sample_to_source_distance() / 1000  # (m)
        lambda_low = wavelength_from_tof(distance=distance, tof=tof_low)
        lambda_high = wavelength_from_tof(distance=distance, tof=tof_high)
        return (round(lambda_low, 2), round(lambda_high, 2))

    def _tof_to_lambda(self, tof):
        """given tof in s, return lambda in Angstrom"""
        neutron_mass = 1.67492749804e-27
        h = 6.62607015e-34
        distance = self._get_sample_to_source_distance() / 1000  # convert to m
        return h / (neutron_mass * (distance / tof)) * 1e10

    def _get_sample_to_source_distance(self) -> float:
        """get sample to source distance in mm"""
        try:
            dist = abs(self._nxs_file["entry/instrument/source/distance"][...]) * 1000
            return dist
        except (KeyError, ValueError):
            logger.warning("sample to moderator_distance not found, using dummy value")
            return 156714

    def _get_panel_gain(self) -> float:
        return 1.0

    def get_goniometer_phi_angle(self) -> float:
        return self.get_goniometer_orientations()[1]

    def get_goniometer(self, index: int = None) -> Goniometer:
        rotation_axis = (0.0, 1.0, 0.0)
        fixed_rotation = (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
        goniometer = GoniometerFactory.make_goniometer(rotation_axis, fixed_rotation)
        try:
            angles = self.get_goniometer_orientations()
        except KeyError:
            logger.warning("crystal_rotation not found, using default")
            return goniometer
        axes = ((1, 0, 0), (0, 1, 0), (0, 0, 1))
        for idx, angle in enumerate(angles):
            goniometer.rotate_around_origin(axes[idx], angle)
        return goniometer

    def get_goniometer_orientations(self) -> tuple[float, float, float]:
        # Angles in deg along x, y, z
        return self._nxs_file["entry/sample/crystal_rotation"][...]

    def get_scan(self, index=None) -> Scan:
        image_range = (1, self.get_num_images())
        properties = {"time_of_flight": tuple(self._get_time_of_flight())}
        return ScanFactory.make_scan_from_properties(
            image_range=image_range, properties=properties
        )

    def get_proton_charge(self) -> float:
        """McStas Simulations don't have a proton charge
        so this is a calculated value"""
        return self._nxs_file["entry/metadata/mcstas_weight2count_scale_factor"][
            ...
        ].item()


class FormatESSNMX_McStas(FormatHDF5):
    """
    Class to read files from NMX
    https://europeanspallationsource.se/instruments/nmx
    preprocessed files in scipp to obtain binned data

    Simulated data coming from McStas.
    """

    def __init__(self, image_file, **kwargs) -> None:
        if not FormatESSNMX_McStas.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self._nxs_file = h5py.File(image_file, "r")
        self._raw_data = None

    @staticmethod
    def understand(image_file: str) -> bool:
        try:
            return FormatESSNMX_McStas.is_nmx_file(image_file)
        except (OSError, KeyError):
            return False

    @staticmethod
    def is_nmx_file(image_file: str) -> bool:
        def get_name(image_file):
            try:
                with h5py.File(image_file, "r") as handle:
                    return handle["entry/instrument/name"][...].item().decode()
            except (OSError, KeyError, AttributeError):
                return ""

        def is_McStas(image_file: str) -> bool:
            try:
                with h5py.File(image_file, "r") as handle:
                    if (
                        handle.get(
                            "entry/metadata/mcstas_weight2count_scale_factor", None
                        )
                        is not None
                    ):
                        return True
                    return False
            except (OSError, KeyError, AttributeError):
                return False

        return ("NMX" in get_name(image_file)) and (is_McStas(image_file))

    def get_instrument_name(self) -> str:
        return "NMX"

    def get_experiment_description(self) -> str:
        return "Simulated data"

    def _load_raw_data(self) -> None:
        raw_data = []
        for panel in self._get_panels():
            spectra = panel["data"][...]
            raw_data.append(flumpy.from_numpy(np.ascontiguousarray(spectra)))

        self._raw_data = tuple(raw_data)

    def get_raw_data(
        self, index: int, use_loaded_data: bool = False
    ) -> tuple[flex.int]:
        raw_data = []
        # image_size = self._get_image_size()
        # total_pixels = image_size[0] * image_size[1]

        if use_loaded_data:
            if self._raw_data is None:
                self._load_raw_data()
            for panel in self._raw_data:
                data = panel[:, :, index : index + 1].astype(np.int32)
                data.reshape(flex.grid(panel.all()[0], panel.all()[1]))
                data.matrix_transpose_in_place()
                raw_data.append(data)

        else:
            for panel in self._get_panels():
                spectra = panel["data"][:, :, index].astype(np.int32)
                raw_data.append(flumpy.from_numpy(np.ascontiguousarray(spectra)))

        return tuple(raw_data)

    def _get_time_channel_bins(self) -> list[float]:
        # (usec)
        # the tofs are recorded separately per panel but they
        # should all be the same
        for panel in self._get_panels():
            return panel["time_of_flight"][...] * 1e6

    def _get_time_of_flight(self) -> list[float]:
        # (usec)
        bins = self._get_time_channel_bins()
        return [float((bins[i] + bins[i + 1]) * 0.5) for i in range(len(bins) - 1)]

    def get_num_images(self) -> int:
        return len(self._get_time_of_flight())

    def get_detector(self, index: int = None) -> Detector:
        panel_names = self._get_panel_names()
        panel_type = self._get_panel_type()
        trusted_range = self._get_panel_trusted_range()
        pixel_size = self._get_pixel_size()
        gain = self._get_panel_gain()
        detector = Detector()
        root = detector.hierarchy()
        panels = self._get_panels()
        panel_projections = self._get_panel_projections_2d(panels)
        for panel_dset, panel_name in zip(panels, panel_names):
            panel = root.add_panel()
            panel.set_type(panel_type)
            panel.set_name(panel_name)
            panel.set_image_size(
                self._get_image_size(panel_dset)
            )  # XXX fix to include detectors possibly not being the same size
            panel.set_trusted_range(trusted_range)
            panel.set_pixel_size(pixel_size)
            fast_axis = self._get_panel_fast_axes(panel_dset)
            slow_axis = self._get_panel_slow_axes(panel_dset)
            panel_origin = self._get_panel_origins(panel_dset)
            panel.set_local_frame(fast_axis, slow_axis, panel_origin)

            panel.set_gain(gain)
            i = int(panel_name[-1])
            r, t = panel_projections[i]
            r = tuple(map(int, r))
            t = tuple(map(int, t))
            panel.set_projection_2d(r, t)

        return detector

    def _get_num_panels(self) -> int:
        return len(self._get_panels())

    def _get_panels(self) -> list[h5py._hl.group.Group]:
        """get the detector panel locations in file"""
        panels = []
        inst_dset = self._nxs_file["/entry/instrument/"]
        for _, dset in inst_dset.items():
            if dset.attrs.get("NX_class") == "NXdetector":
                panels.append(dset)
        return panels

    def _get_panel_names(self) -> list[str]:
        panel_names = []
        inst_dset = self._nxs_file["/entry/instrument/"]
        for name, dset in inst_dset.items():
            if dset.attrs.get("NX_class") == "NXdetector":
                panel_names.append(name)
        return panel_names

    def _get_panel_name(self, panel) -> str:
        return panel.name.split("/")[-1]

    def _get_panel_type(self) -> str:
        return "Triple_GEM_Gd"

    def _get_image_size(self, panel) -> tuple[int, int]:
        # (px)
        dset = panel["data"]
        return dset[:, :, 0].shape

    def _get_panel_trusted_range(self) -> tuple[int, int]:
        # 4 * 1280**2 plus buffer
        return (-1, 2**64 - 1)

    def _get_pixel_size(self) -> tuple[float, float]:
        # (mm)
        return (0.4, 0.4)

    def _get_panel_fast_axes(self, panel) -> tuple[float, float, float]:
        # return ((1.0, 0.0, 0.0), (0.0, 0.0, 1.0), (0.0, 0.0, -1.0))
        fast_axis = panel["fast_axis"][...]
        return tuple(fast_axis)

    def _get_panel_slow_axes(self, panel) -> tuple[float, float, float]:
        # return ((0.0, 1.0, 0.0), (0.0, 1.0, 0.0), (0.0, 1.0, 0.0))
        slow_axis = panel["slow_axis"][...]
        return tuple(slow_axis)

    def _get_panel_origins(self, panel) -> tuple[float, float, float]:
        # (mm)
        # return ((-250, -250.0, -292.0), (290, -250.0, -250), (-290, -250.0, 250.0))

        origin = panel["origin"][...]
        panelnum = int(panel.name[-1])
        origin *= 1000  # convert to mm
        corrfact = self._get_correction_factor()[panelnum]

        corrorg = origin + corrfact
        return tuple(corrorg)

    def _get_correction_factor(self):
        return np.array([[0, -256, -256], [-256.0, -256.0, 0], [0, -256, 256]])

    def _get_panel_projections_2d(self, panels) -> dict[int : tuple[tuple, tuple]]:
        p_w, p_h = self._get_image_size(panels[0])  # XXX fix later
        p_w += 10
        p_h += 10
        panel_pos = {
            int(panels[0].name[-1]): ((-1, 0, 0, -1), (p_h, 0)),
            int(panels[1].name[-1]): ((-1, 0, 0, -1), (p_h, p_w)),
            int(panels[2].name[-1]): ((-1, 0, 0, -1), (p_h, -p_w)),
        }

        return panel_pos

    def get_beam(self, index: int = None) -> PolychromaticBeam:
        direction = self._get_sample_to_source_direction()
        distance = self._get_sample_to_source_distance()
        wavelength_range = self._get_wavelength_range()
        return BeamFactory.make_polychromatic_beam(
            direction=direction,
            sample_to_source_distance=distance,
            probe=Probe.neutron,
            wavelength_range=wavelength_range,
        )

    def _get_sample_to_source_direction(self) -> tuple[float, float, float]:
        return (0, 0, -1)

    def _get_wavelength_range(self) -> tuple[float, float]:
        # (A)
        tofs = np.array(self._get_time_of_flight()) / 1e6
        tof_low = tofs[0]
        tof_high = tofs[-1]
        lambda_low = self._tof_to_lambda(tof_low)
        lambda_high = self._tof_to_lambda(tof_high)
        return (round(lambda_low, 2), round(lambda_high, 2))

    def _tof_to_lambda(self, tof):
        """given tof in s, return lambda in Angstrom"""
        neutron_mass = 1.67492749804e-27
        h = 6.62607015e-34
        distance = self._get_sample_to_source_distance() / 1000  # convert to m
        return h / (neutron_mass * (distance / tof)) * 1e10

    def _get_sample_to_source_distance(self) -> float:
        """get sample to source distance in mm"""
        try:
            dist = abs(self._nxs_file["entry/instrument/source/distance"][...]) * 1000
            return dist
        except (KeyError, ValueError):
            logger.warning("sample to moderator_distance not found, using dummy value")
            return 156714

    def _get_panel_gain(self) -> float:
        return 1.0

    def get_goniometer_phi_angle(self) -> float:
        return self.get_goniometer_orientations()[1]

    def get_goniometer(self, index: int = None) -> Goniometer:
        rotation_axis = (0.0, 1.0, 0.0)
        fixed_rotation = (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
        goniometer = GoniometerFactory.make_goniometer(rotation_axis, fixed_rotation)
        try:
            angles = self.get_goniometer_orientations()
        except KeyError:
            logger.warning("crystal_rotation not found, using default")
            return goniometer
        axes = ((1, 0, 0), (0, 1, 0), (0, 0, 1))
        for idx, angle in enumerate(angles):
            goniometer.rotate_around_origin(axes[idx], angle)
        return goniometer

    def get_goniometer_orientations(self) -> tuple[float, float, float]:
        # Angles in deg along x, y, z
        return self._nxs_file["entry/sample/crystal_rotation"][...]

    def get_scan(self, index=None) -> Scan:
        image_range = (1, self.get_num_images())
        properties = {"time_of_flight": tuple(self._get_time_of_flight())}
        return ScanFactory.make_scan_from_properties(
            image_range=image_range, properties=properties
        )

    def get_proton_charge(self) -> float:
        """McStas Simulations don't have a proton charge
        so this is a calculated value"""
        return self._nxs_file["entry/metadata/mcstas_weight2count_scale_factor"][
            ...
        ].item()



"""
An implementation of the CBF image reader for ADSC images, from the ADSC
HF-4M SN H401 currently on APS sector 24 (NE-CAT).
Located in dxtbx/format
"""

from __future__ import annotations

import binascii
import calendar
import re
import sys
import time

from cctbx.eltbx import attenuation_coefficient
from iotbx.detectors.adsc_minicbf import ADSCHF4MImage

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFMini import FormatCBFMini
from dxtbx.model import ParallaxCorrectedPxMmStrategy


def get_adsc_timestamp(timestamp):
    # reduce duplicate to single underscore
    timestamp = re.sub("_+", "_", timestamp)

    for format in ["%a_%b_%d_%H:%M:%S_%Y"]:
        try:
            struct_time = time.strptime(timestamp, format)
            return calendar.timegm(struct_time)

        except Exception:
            pass

    raise RuntimeError("timestamp %s not recognised" % timestamp)


class FormatCBFMiniADSCHF4M(FormatCBFMini):
    """A class for reading mini CBF format ADSC images for HF-4M @ NE-CAT."""

    @staticmethod
    def understand(image_file):
        header = FormatCBFMini.get_cbf_header(image_file)

        for record in header.split("\n"):
            if "# Detector" in record and "ADSC" in record and "HF-4M" in header:
                return True
            if "_array_data.header_convention" in record and "ADSC" in record:
                return True

        return False

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._cif_header_dictionary["Detector_distance"].split()[0])

        beam_xy = (
            self._cif_header_dictionary["Beam_xy"]
            .replace("(", "")
            .replace(")", "")
            .replace(",", "")
            .split()[:2]
        )

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        beam_x, beam_y = map(float, beam_xy)

        pixel_xy = (
            self._cif_header_dictionary["Pixel_size"]
            .replace("m", "")
            .replace("x", "")
            .split()
        )

        pixel_x, pixel_y = map(float, pixel_xy)

        thickness = float(self._cif_header_dictionary["Silicon"].split()[2]) * 1000.0

        nx = int(self._cif_header_dictionary["X-Binary-Size-Fastest-Dimension"])
        ny = int(self._cif_header_dictionary["X-Binary-Size-Second-Dimension"])

        overload = int(self._cif_header_dictionary["Count_cutoff"].split()[0])
        minimum_trusted_value = 0

        # take into consideration here the thickness of the sensor also the
        # wavelength of the radiation (which we have in the same file...)
        table = attenuation_coefficient.get_table("Si")
        mu = table.mu_at_angstrom(wavelength) / 10.0
        t0 = thickness

        # FIXME would also be very nice to be able to take into account the
        # misalignment of the individual modules given the calibration information...

        detector = self._detector_factory.simple(
            "PAD",
            distance * 1000.0,
            (beam_x * pixel_x * 1000.0, beam_y * pixel_y * 1000.0),
            "+x",
            "-y",
            (1000 * pixel_x, 1000 * pixel_y),
            (nx, ny),
            (minimum_trusted_value, overload),
            [],
            ParallaxCorrectedPxMmStrategy(mu, t0),
        )
        """
    for f0, f1, s0, s1 in determine_pilatus_mask(detector):
      detector[0].add_mask(f0-1, s0-1, f1, s1)
    """
        return detector

    def _beam(self):
        """Return a simple model for the beam."""

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        beam = self._beam_factory.simple(wavelength)

        try:
            flux = float(self._cif_header_dictionary["Flux"].split()[0])
            beam.set_flux(flux)
        except KeyError:
            pass

        try:
            transmission = float(self._cif_header_dictionary["Transmission"].split()[0])
            beam.set_transmission(transmission)
        except KeyError:
            pass

        return beam

    def _scan(self):
        """Return the scan information for this image."""

        exposure_time = float(self._cif_header_dictionary["Exposure_period"].split()[0])

        osc_start = float(self._cif_header_dictionary["Start_angle"].split()[0])
        osc_range = float(self._cif_header_dictionary["Angle_increment"].split()[0])

        timestamp = get_adsc_timestamp(self._cif_header_dictionary["timestamp"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, timestamp
        )

    def detectorbase_start(self):
        self.detectorbase = ADSCHF4MImage(self._image_file)
        self.detectorbase.readHeader()

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values

    def get_raw_data(self):
        if self._raw_data is None:
            data = self._read_cbf_image()
            self._raw_data = data

        return self._raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniADSCHF4M.understand(arg))



from __future__ import annotations

from pathlib import Path

import h5py
import pytest

from dxtbx.format import nexus

# Some names for sample NXentries in a dummy NeXus file.
entry_names = "/entry", "/another_entry", "/yet_another_entry"


@pytest.fixture(scope="session")
def nexus_file(tmp_path_factory) -> Path:
    """
    Create a dummy NXmx-flavoured NeXus HDF5 file to exercise parts of format/nexus.py.

    The file contains
        - Three NXentries.
        - Two NXsamples within the NXentry 'entry'.
        - A NXinstrument within 'entry'.
        - A NXbeam within the NXsample 'sample', with a soft link from the NXinstrument.
        - A broken soft link from the NXinstrument.
        - A HDF5 group and a data set without NXclass attributes, both within 'entry'.

    Args:
        tmp_path_factory:  Pytest session-scoped temporary directory fixture.

    Returns:
        Path to the dummy NeXus file.
    """
    h5_file = tmp_path_factory.mktemp("dxtbx_dummy_nexus", numbered=False) / "tmp.nxs"

    with h5py.File(h5_file, "w") as fw:
        # Create some top-level NXmx NXentries.
        for name in entry_names:
            group = fw.create_group(name)
            group.attrs["NX_class"] = "NXentry"
            group["definition"] = "NXmx"
        entry = fw["entry"]

        # Create two NXsamples in the NXentry entry.
        for name in "sample", "another_sample":
            dataset = entry.create_group(name)
            dataset.attrs["NX_class"] = "NXsample"
        sample = entry["sample"]

        # Create a NXinstrument in the NXentry entry.
        instrument = entry.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"

        # Create a NXbeam in the NXsample sample.
        beam = sample.create_group("beam")
        beam.attrs["NX_class"] = "NXbeam"
        # Link to it from the NXinstrument instrument.
        instrument["beam"] = h5py.SoftLink(beam.name)

        # Ensure that find_classes gracefully handles broken soft/external links
        instrument["broken_link"] = h5py.SoftLink("/entry/nonsense")

        # Ensure that find_classes copes with non-NX_class groups and data sets.
        entry.create_group("group")
        entry.create_dataset("dataset", data="Test data set.")

    return h5_file


def test_data_factory_nxs(dials_data):
    nxs_file = dials_data("vmxi_thaumatin") / "image_15799.nxs"
    with h5py.File(nxs_file) as fh:
        data = fh["/entry/data"]
        data_factory = nexus.DataFactory(nexus.NXdata(data))
        for i in (0, data["data"].shape[0] - 1):
            assert data_factory[i].size()
        # If we try and access an image beyond the size of the dataset it should fail.
        # See https://github.com/cctbx/dxtbx/issues/285
        with pytest.raises(IndexError):
            data_factory[data["data"].shape[0]]


def test_find_entries(nexus_file):
    with h5py.File(nexus_file, "r") as fr:
        entries = nexus.find_entries(fr)
        assert len(entries) == len(entry_names)
        assert {entry.name for entry in entries} == set(entry_names)


def test_find_entries_no_entry(tmp_path):
    h5_file = tmp_path / "tmp.nxs"
    with h5py.File(h5_file, "w") as fw:
        # NX_class == NXdata
        group = fw.create_dataset("data", (100,))
        group.attrs["NX_class"] = "NXdata"
        # NX_class == NXentry but no definition provided - skip this entry
        group = fw.create_group("no_definition")
        group.attrs["NX_class"] = "NXentry"
        # NX_class == NXentry but definition != NXmx
        group = fw.create_group("wrong_definition")
        group.attrs["NX_class"] = "NXentry"
        group["definition"] = "NXmagic"

    with h5py.File(h5_file, "r") as fr:
        entries = nexus.find_entries(fr)
        assert len(entries) == 0


def test_find_entries_empty_file(tmp_path):
    h5_file = tmp_path / "tmp.nxs"
    with h5py.File(h5_file, "w"):
        pass

    with h5py.File(h5_file, "r") as fr:
        entries = nexus.find_entries(fr)
        assert len(entries) == 0


def test_find_classes(nexus_file):
    """
    Test dxtbx.format.nexus.find_classes.

    1. Call find_classes on the main NXentry of the dummy NeXus file.
    2. Ensure it finds one instrument, two samples and three children that do not
       belong to a NX_class.
    3. Ensure that the children that don't belong to a NX_class are 'definition',
       'group' and 'dataset'.
    """
    with h5py.File(nexus_file, "r") as fr:
        samples, instruments, classless = nexus.find_classes(
            fr["entry"], "NXsample", "NXinstrument", None
        )
        assert len(instruments) == 1
        assert len(samples) == 2
        assert len(classless) == 3
        expected_classless = {"/entry/definition", "/entry/group", "/entry/dataset"}
        assert {value.name for value in classless} == expected_classless


def test_find_class(nexus_file):
    """
    Test dxtbx.format.nexus.find_class

    1. Ensure that nexus.find_class finds the single NXbeam child of /entry/sample.
    2. Ensure that the broken link /entry/instrument/broken_link exists but is
       ignored by nexus.find_class, even in a search for children without an NX_class.
    """
    with h5py.File(nexus_file, "r") as fr:
        beams = nexus.find_class(fr["entry/sample"], "NXbeam")
        assert len(beams) == 1
        assert beams[0].name == "/entry/sample/beam"

        instrument = fr["entry/instrument"]
        classless = nexus.find_class(instrument, None)
        assert "broken_link" in instrument and not classless


def test_find_class_softlink(nexus_file):
    """
    Test that nexus.find_class follows soft links.

    Call find_class on /entry/instrument and assert that it correctly follows the
    soft link to /entry/sample/beam.
    """
    with h5py.File(nexus_file, "r") as fr:
        assert len(nexus.find_class(fr["entry/instrument"], "NXbeam")) == 1
        assert len(nexus.find_class(fr["entry/sample"], "NXbeam")) == 1



from __future__ import annotations

import pickle
from copy import deepcopy
from math import sqrt

import pytest

from libtbx.phil import parse
from scitbx import matrix

from dxtbx.model import Beam, Detector
from dxtbx.model.detector import (
    DetectorFactory,
    ParallaxCorrectedPxMmStrategy,
    detector_phil_scope,
)


@pytest.fixture
def detector():
    detector = Detector()
    root = detector.hierarchy()
    root.set_name("D1")
    root.set_type("D")

    quad1 = root.add_group()
    quad1.set_name("Q1")
    quad1.set_type("Q")
    panel1 = quad1.add_panel()
    panel1.set_name("P1")
    panel1.set_type("P")
    panel2 = quad1.add_panel()
    panel2.set_name("P2")
    panel2.set_type("P")

    quad2 = root.add_group()
    quad2.set_name("Q2")
    quad2.set_type("Q")
    panel3 = quad2.add_panel()
    panel3.set_name("P3")
    panel3.set_type("P")
    panel4 = quad2.add_panel()
    panel4.set_name("P4")
    panel4.set_type("P")

    return detector


def test_flat(detector):
    """Test the flat hierarchy."""

    expected_types = ["P", "P", "P", "P"]
    expected_names = ["P1", "P2", "P3", "P4"]
    names = []
    types = []
    for p in detector:
        names.append(p.get_name())
        types.append(p.get_type())
    assert all(n == en for n, en in zip(names, expected_names))
    assert all(t == et for t, et in zip(types, expected_types))


def test_child_iteration(detector):
    # Iterate through the detector's children and check output
    expected_types = ["Q", "Q"]
    expected_names = ["Q1", "Q2"]
    names = []
    types = []
    for p in detector.hierarchy():
        names.append(p.get_name())
        types.append(p.get_type())
    assert all(n == en for n, en in zip(names, expected_names))
    assert all(t == et for t, et in zip(types, expected_types))


def test_child_iteration_in_reverse(detector):
    # Iterate through the detector's children in reverse and check output
    expected_types = ["Q", "Q"]
    expected_names = ["Q2", "Q1"]
    names = []
    types = []
    for p in reversed(detector.hierarchy()):
        names.append(p.get_name())
        types.append(p.get_type())
    assert all(n == en for n, en in zip(names, expected_names))
    assert all(t == et for t, et in zip(types, expected_types))


def test_child_indexing(detector):
    # Use an index to access the detector children
    assert len(detector.hierarchy()) == 2
    group = detector.hierarchy()[1]
    assert group.get_name() == "Q2" and group.get_type() == "Q"
    assert len(group) == 2
    panel = group[0]
    assert panel.get_name() == "P3" and panel.get_type() == "P"


def test_depth_first_iteration(detector):
    # Iterate through the tree pre-order and check output
    expected_types = ["D", "Q", "P", "P", "Q", "P", "P"]
    expected_names = ["D1", "Q1", "P1", "P2", "Q2", "P3", "P4"]
    names = []
    types = []
    for p in detector.iter_preorder():
        names.append(p.get_name())
        types.append(p.get_type())
    assert all(n == en for n, en in zip(names, expected_names))
    assert all(t == et for t, et in zip(types, expected_types))


def test_breadth_first_iteration(detector):
    # Iterate through the tree level-order and check output
    expected_types = ["D", "Q", "Q", "P", "P", "P", "P"]
    expected_names = ["D1", "Q1", "Q2", "P1", "P2", "P3", "P4"]
    names = []
    types = []
    for p in detector.iter_levelorder():
        names.append(p.get_name())
        types.append(p.get_type())
    assert all(n == en for n, en in zip(names, expected_names))
    assert all(t == et for t, et in zip(types, expected_types))


def test_panels_depth_first_iteration(detector):
    # Iterate through the panels in pre-order and check output
    expected_types = ["P", "P", "P", "P"]
    expected_names = ["P1", "P2", "P3", "P4"]
    names = []
    types = []
    for p in detector.iter_panels():
        names.append(p.get_name())
        types.append(p.get_type())
    assert all(n == en for n, en in zip(names, expected_names))
    assert all(t == et for t, et in zip(types, expected_types))


def test_get_valid_D_matrix(detector):
    """Setup the hierarchy of frames and check it's all consistent."""
    # Set a valid frame for the top level detector
    detector.hierarchy().set_local_frame(
        (1, 0, 0),
        (0, 1, 0),
        (0, 0, 100),  # Fast axis  # Slow axis
    )  # Origin

    # Check that all sub groups have the same frame and that we can get
    # a valid D matrix
    for obj in detector.iter_preorder():
        fast = matrix.col(obj.get_fast_axis())
        slow = matrix.col(obj.get_slow_axis())
        orig = matrix.col(obj.get_origin())
        assert abs(fast - matrix.col((1, 0, 0))) < 1e-7
        assert abs(slow - matrix.col((0, 1, 0))) < 1e-7
        assert abs(orig - matrix.col((0, 0, 100))) < 1e-7
        assert obj.get_D_matrix()

    # Get the quadrants and set their frames
    q1, q2 = detector.hierarchy().children()
    q1.set_local_frame(
        (1, 1, 0),  # Fast axis relative to detector frame
        (-1, 1, 0),  # Slow axis relative to detector frame
        (10, 10, 0),
    )  # Origin relative to detector frame
    q2.set_local_frame(
        (1, -1, 0),  # Fast axis relative to detector frame
        (1, 1, 0),  # Slow axis relative to detector frame
        (20, 20, 0),
    )  # Origin relative to detector frame

    # Get the panels and set their frames
    p1, p2 = q1.children()
    p1.set_local_frame(
        (1, -1, 0),  # Fast axis relative to q1 frame
        (1, 1, 0),  # Slow axis relative to q1 frame
        (5, 0, 10),
    )  # Origin relative to q1 frame
    p2.set_local_frame(
        (1, 1, 0),  # Fast axis relative to q1 frame
        (-1, 1, 0),  # Slow axis relative to q1 frame
        (0, 5, -10),
    )  # Origin relative to q1 frame

    # Get the panels and set their frames
    p3, p4 = q2.children()
    p3.set_local_frame(
        (1, -1, 0),  # Fast axis relative to q2 frame
        (1, 1, 0),  # Slow axis relative to q2 frame
        (0, 5, -10),
    )  # Origin relative to q2 frame
    p4.set_local_frame(
        (1, 1, 0),  # Fast axis relative to q2 frame
        (-1, 1, 0),  # Slow axis relative to q2 frame
        (5, 0, 10),
    )  # Origin relative to q2 frame

    # Test the panel coordinate systems

    eps = 1e-7
    p1_d0 = matrix.col((10.0 + sqrt(5.0**2 / 2), 10.0 + sqrt(5.0**2 / 2), 110))
    p2_d0 = matrix.col((10.0 - sqrt(5.0**2 / 2), 10.0 + sqrt(5.0**2 / 2), 90))
    p3_d0 = matrix.col((20.0 + sqrt(5.0**2 / 2), 20.0 + sqrt(5.0**2 / 2), 90))
    p4_d0 = matrix.col((20.0 + sqrt(5.0**2 / 2), 20.0 - sqrt(5.0**2 / 2), 110))
    p1_d1 = matrix.col((1, 0, 0))
    p2_d1 = matrix.col((0, 1, 0))
    p3_d1 = matrix.col((0, -1, 0))
    p4_d1 = matrix.col((1, 0, 0))
    p1_d2 = matrix.col((0, 1, 0))
    p2_d2 = matrix.col((-1, 0, 0))
    p3_d2 = matrix.col((1, 0, 0))
    p4_d2 = matrix.col((0, 1, 0))
    assert abs(matrix.col(p1.get_origin()) - p1_d0) < eps
    assert abs(matrix.col(p2.get_origin()) - p2_d0) < eps
    assert abs(matrix.col(p3.get_origin()) - p3_d0) < eps
    assert abs(matrix.col(p4.get_origin()) - p4_d0) < eps
    assert abs(matrix.col(p1.get_fast_axis()) - p1_d1) < eps
    assert abs(matrix.col(p2.get_fast_axis()) - p2_d1) < eps
    assert abs(matrix.col(p3.get_fast_axis()) - p3_d1) < eps
    assert abs(matrix.col(p4.get_fast_axis()) - p4_d1) < eps
    assert abs(matrix.col(p1.get_slow_axis()) - p1_d2) < eps
    assert abs(matrix.col(p2.get_slow_axis()) - p2_d2) < eps
    assert abs(matrix.col(p3.get_slow_axis()) - p3_d2) < eps
    assert abs(matrix.col(p4.get_slow_axis()) - p4_d2) < eps


def test_copy_and_reference(detector):
    # Get the detector hierarchy
    root = detector.hierarchy()

    # Get the panels in the hierarchy
    p1 = root[0][0]
    p2 = root[0][1]
    p3 = root[1][0]
    p4 = root[1][1]

    # Check panels are the same
    assert p1.is_(detector[0])
    assert p2.is_(detector[1])
    assert p3.is_(detector[2])
    assert p4.is_(detector[3])

    # Copy the detector
    new_detector = deepcopy(detector)

    # Check they're the same
    assert new_detector == detector

    # Add an offset to propagate
    root = new_detector.hierarchy()
    root.set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 10))

    # Get the panels in the hierarchy
    p1 = root[0][0]
    p2 = root[0][1]
    p3 = root[1][0]
    p4 = root[1][1]

    # Check panels are the same
    assert p1.is_(new_detector[0])
    assert p2.is_(new_detector[1])
    assert p3.is_(new_detector[2])
    assert p4.is_(new_detector[3])


def test_pickle(detector):
    # Get the detector hierarchy
    root = detector.hierarchy()

    # Get the panels in the hierarchy
    p1 = root[0][0]
    p2 = root[0][1]
    p3 = root[1][0]
    p4 = root[1][1]

    # Check panels are the same
    assert p1.is_(detector[0])
    assert p2.is_(detector[1])
    assert p3.is_(detector[2])
    assert p4.is_(detector[3])

    # Copy the detector
    new_detector = pickle.loads(pickle.dumps(detector))

    # Check they're the same
    assert new_detector == detector

    # Add an offset to propagate
    root = new_detector.hierarchy()
    root.set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 10))

    # Get the panels in the hierarchy
    p1 = root[0][0]
    p2 = root[0][1]
    p3 = root[1][0]
    p4 = root[1][1]

    # Check panels are the same
    assert p1.is_(new_detector[0])
    assert p2.is_(new_detector[1])
    assert p3.is_(new_detector[2])
    assert p4.is_(new_detector[3])


def test_from_phil():
    beam = Beam((0, 0, 1))

    params = detector_phil_scope.fetch(
        parse(
            """
    detector {
      panel {
        id = 0
        origin = (1, 1, 1)
        pixel_size = (0.001,0.001)
        image_size = (1000,1000)
        trusted_range = (-1, 1000)
        material = "Si"
        thickness = 0.01
        parallax_correction = True
      }
      panel {
        id = 1
        origin = (2, 2, 2)
        pixel_size = (0.001,0.001)
        image_size = (1000,1000)
        trusted_range = (-1, 1000)
      }
      panel {
        id = 2
        origin = (3, 3, 3)
        pixel_size = (0.001,0.001)
        image_size = (1000,1000)
        trusted_range = (-1, 1000)
      }
      panel {
        id = 3
        origin = (4, 4, 4)
        pixel_size = (0.001,0.001)
        image_size = (1000,1000)
        trusted_range = (-1, 1000)
      }

      hierarchy {
        name = "Root"
        origin = (100, 100, 100)
        group {
          id = 0
          origin = (10, 10, 10)
        }
        group {
          id = 0,0
          origin = (1, 1, 1)
          panel = 0
        }
        group {
          id = 0,1
          origin = (2, 2, 2)
          panel = 1
        }
        group {
          id = 1
          origin = (20, 20, 20)
        }
        group {
          id = 1,0
          origin = (1, 1, 1)
          panel = 2
        }
        group {
          id = 1,1
          origin = (2, 2, 2)
          panel = 3
        }
      }
    }
  """
        )
    ).extract()

    # Test create model
    d1 = DetectorFactory.from_phil(params, beam=beam)

    root = d1.hierarchy()

    # Check hierarchy origins
    assert root.get_origin() == (100, 100, 100)
    assert root[0].get_origin() == (110, 110, 110)
    assert root[1].get_origin() == (120, 120, 120)
    assert root[0][0].get_origin() == (111, 111, 111)
    assert root[0][1].get_origin() == (112, 112, 112)
    assert root[1][0].get_origin() == (121, 121, 121)
    assert root[1][1].get_origin() == (122, 122, 122)
    assert root[0][0][0].get_origin() == (112, 112, 112)
    assert root[0][1][0].get_origin() == (114, 114, 114)
    assert root[1][0][0].get_origin() == (124, 124, 124)
    assert root[1][1][0].get_origin() == (126, 126, 126)

    # Check panels are correct in hierarchy
    assert root[0][0][0].is_(d1[0])
    assert root[0][1][0].is_(d1[1])
    assert root[1][0][0].is_(d1[2])
    assert root[1][1][0].is_(d1[3])

    # Check panel attributes
    assert d1[0].get_image_size() == (1000, 1000)
    assert d1[0].get_pixel_size() == (0.001, 0.001)
    assert d1[0].get_trusted_range() == (-1, 1000)
    assert d1[0].get_material() == "Si"
    assert d1[0].get_thickness() == 0.01
    assert isinstance(d1[0].get_px_mm_strategy(), ParallaxCorrectedPxMmStrategy)

    assert d1[1].get_image_size() == (1000, 1000)
    assert d1[1].get_pixel_size() == (0.001, 0.001)
    assert d1[1].get_trusted_range() == (-1, 1000)

    assert d1[2].get_image_size() == (1000, 1000)
    assert d1[2].get_pixel_size() == (0.001, 0.001)
    assert d1[2].get_trusted_range() == (-1, 1000)

    assert d1[3].get_image_size() == (1000, 1000)
    assert d1[3].get_pixel_size() == (0.001, 0.001)
    assert d1[3].get_trusted_range() == (-1, 1000)

    params = detector_phil_scope.fetch(
        parse(
            """
    detector {
      panel {
        id = 0
        parallax_correction = False
      }
      panel {
        id = 1
        material = "Si"
        thickness = 0.01
        parallax_correction = True
      }

      hierarchy {
        name = "Root"
        origin = (200, 200, 200)
        group {
          id = 0
          origin = (20, 20, 20)
        }
        group {
          id = 0,0
          origin = (2, 2, 2)
        }
        group {
          id = 0,1
          origin = (3, 3, 3)
        }
        group {
          id = 1
          origin = (30, 30, 30)
        }
        group {
          id = 1,0
          origin = (2, 2, 2)
        }
        group {
          id = 1,1
          origin = (3, 3, 3)
        }
      }
    }
  """
        )
    ).extract()

    # Test overwrite model
    d2 = DetectorFactory.from_phil(params, reference=d1, beam=beam)

    root = d2.hierarchy()

    # Check hierarchy origins
    assert root.get_origin() == (200, 200, 200)
    assert root[0].get_origin() == (220, 220, 220)
    assert root[1].get_origin() == (230, 230, 230)
    assert root[0][0].get_origin() == (222, 222, 222)
    assert root[0][1].get_origin() == (223, 223, 223)
    assert root[1][0].get_origin() == (232, 232, 232)
    assert root[1][1].get_origin() == (233, 233, 233)
    assert root[0][0][0].get_origin() == (223, 223, 223)
    assert root[0][1][0].get_origin() == (225, 225, 225)
    assert root[1][0][0].get_origin() == (235, 235, 235)
    assert root[1][1][0].get_origin() == (237, 237, 237)

    # Check panels are correct in hierarchy
    assert root[0][0][0].is_(d2[0])
    assert root[0][1][0].is_(d2[1])
    assert root[1][0][0].is_(d2[2])
    assert root[1][1][0].is_(d2[3])

    # Check panel attributes
    assert not isinstance(d2[0].get_px_mm_strategy(), ParallaxCorrectedPxMmStrategy)
    assert d2[1].get_material() == "Si"
    assert d2[1].get_thickness() == 0.01
    assert isinstance(d2[1].get_px_mm_strategy(), ParallaxCorrectedPxMmStrategy)



"""
An implementation of the SMV image reader for Rigaku Saturn images.
Inherits from FormatSMVRigaku.
"""

from __future__ import annotations

import sys
import time

from iotbx.detectors.noir import NoirImage
from scitbx import matrix

from dxtbx.format.FormatSMVRigaku import FormatSMVRigaku


class FormatSMVNOIR(FormatSMVRigaku):
    """A class for reading SMV format ALS 4.2.2 NOIR images, and correctly
    constructing a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a ALS 4.2.2 NOIR SMV format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMVRigaku.get_smv_header(image_file)

        wanted_header_items = [
            "DETECTOR_NUMBER",
            "DETECTOR_NAMES",
            "CRYSTAL_GONIO_NUM_VALUES",
            "CRYSTAL_GONIO_NAMES",
            "CRYSTAL_GONIO_UNITS",
            "CRYSTAL_GONIO_VALUES",
            "NOIR1_CREATED",
            "ROTATION",
            "ROTATION_AXIS_NAME",
            "ROTATION_VECTOR",
            "SOURCE_VECTORS",
            "SOURCE_WAVELENGTH",
            "SOURCE_POLARZ",
            "DIM",
            "SIZE1",
            "SIZE2",
        ]

        if any(item not in header for item in wanted_header_items):
            return False

        detector_prefix = header["DETECTOR_NAMES"].split()[0].strip()

        more_wanted_header_items = [
            "DETECTOR_DIMENSIONS",
            "DETECTOR_SIZE",
            "DETECTOR_VECTORS",
            "GONIO_NAMES",
            "GONIO_UNITS",
            "GONIO_VALUES",
            "GONIO_VECTORS",
            "SPATIAL_BEAM_POSITION",
        ]

        if any(
            f"{detector_prefix}{item}" not in header
            for item in more_wanted_header_items
        ):
            return False

        return True

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment. Easy from Rigaku Saturn images as
        they contain everything pretty much we need..."""

        super().__init__(image_file, **kwargs)
        self.detector_class = "NOIR1"
        self.detector = "adsc"

    def detectorbase_start(self):
        self.detectorbase = NoirImage(self._image_file)
        self.detectorbase.open_file = self.open_file
        self.detectorbase.readHeader()

    def _goniometer(self):
        """Initialize the structure for the goniometer - this will need to
        correctly compose the axes given in the image header. In this case
        this is made rather straightforward as the image header has the
        calculated rotation axis stored in it. We could work from the
        rest of the header and construct a goniometer model."""

        axis = tuple(map(float, self._header_dictionary["ROTATION_VECTOR"].split()))

        return self._goniometer_factory.known_axis(axis)

    def get_beam_pixels(self, detector_name):
        return [
            float(bp)
            for bp in self._header_dictionary[
                "%sSPATIAL_BEAM_POSITION" % detector_name
            ].split()[:2]
        ]

    def _detector(self):
        """Return a model for the detector, allowing for two-theta offsets
        and the detector position. This will be rather more complex..."""

        detector_name = self._header_dictionary["DETECTOR_NAMES"].split()[0].strip()

        detector_axes = self.get_detector_axes(detector_name)
        detector_fast = matrix.col(tuple(detector_axes[:3]))
        detector_slow = matrix.col(tuple(detector_axes[3:]))

        beam_pixels = self.get_beam_pixels(detector_name)
        pixel_size = self.get_pixel_size(detector_name)
        image_size = self.get_image_size(detector_name)

        detector_origin = -(
            beam_pixels[0] * pixel_size[0] * detector_fast
            + beam_pixels[1] * pixel_size[1] * detector_slow
        )

        gonio_axes = self.get_gonio_axes(detector_name)
        gonio_values = self.get_gonio_values(detector_name)
        gonio_units = self._header_dictionary["%sGONIO_UNITS" % detector_name].split()
        gonio_num_axes = int(
            self._header_dictionary["%sGONIO_NUM_VALUES" % detector_name]
        )

        rotations = []
        translations = []

        for j, unit in enumerate(gonio_units):
            axis = matrix.col(gonio_axes[3 * j : 3 * (j + 1)])
            if unit == "deg":
                rotations.append(
                    axis.axis_and_angle_as_r3_rotation_matrix(gonio_values[j], deg=True)
                )
                translations.append(matrix.col((0.0, 0.0, 0.0)))
            elif unit == "mm":
                rotations.append(
                    matrix.sqr((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
                )
                translations.append(gonio_values[j] * axis)
            else:
                raise RuntimeError("unknown axis unit %s" % unit)

        rotations.reverse()
        translations.reverse()

        for j in range(gonio_num_axes):
            detector_fast = rotations[j] * detector_fast
            detector_slow = rotations[j] * detector_slow
            detector_origin = rotations[j] * detector_origin
            detector_origin = translations[j] + detector_origin

        max_trusted_value = int(float(self._header_dictionary["SATURATED_VALUE"]))
        min_trusted_value = 0

        return self._detector_factory.complex(
            "CCD",
            detector_origin.elems,
            detector_fast.elems,
            detector_slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
        )

    def _beam(self):
        """Return a simple model for the beam."""

        beam_direction = self.get_beam_direction()
        p_fraction, p_plane = self.get_beam_polarization()

        wavelength = float(self._header_dictionary["SOURCE_WAVELENGTH"].split()[-1])

        return self._beam_factory.complex(
            beam_direction, p_fraction, p_plane, wavelength
        )

    def _scan(self):
        """Return the scan information for this image."""
        epoch = time.strptime(
            self._header_dictionary["NOIR1_CREATED"], "%m/%d/%y  %H:%M:%S"
        )
        return self._create_single_SVM_scan(epoch)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVNOIR.understand(arg))



from __future__ import annotations

import pickle

from dxtbx.model import Beam, Detector, Goniometer, Panel, Scan


def pickle_then_unpickle(obj):
    """Pickle to a temp file then un-pickle."""
    return pickle.loads(pickle.dumps(obj))


def test_beam():
    """Test pickling the beam object."""
    obj1 = Beam((1, 1, 1))
    obj2 = pickle_then_unpickle(obj1)
    assert obj1 == obj2


def test_goniometer():
    """Test pickling the goniometer object."""
    obj1 = Goniometer()
    obj2 = pickle_then_unpickle(obj1)
    assert obj1 == obj2


def test_panel():
    """Test pickling the panel object."""
    obj1 = Panel()
    obj1.set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 1))
    obj2 = pickle_then_unpickle(obj1)
    assert obj1 == obj2


def test_detector():
    """Test pickling the detector object."""
    p = Panel()
    p.set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 1))
    obj1 = Detector(p)
    obj2 = pickle_then_unpickle(obj1)
    assert obj1 == obj2


def test_hierarchical_detector():
    """Test pickling the detector object."""
    p = Panel()
    p.set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 1))
    obj1 = Detector()
    root = obj1.hierarchy()
    root.add_panel(p)
    root.add_group()
    obj2 = pickle_then_unpickle(obj1)
    assert obj2.hierarchy()[0] == obj2[0]
    assert obj2.hierarchy()[0] in obj2
    assert obj2.hierarchy()[1].is_group()
    assert obj1 == obj2


def test_scan():
    """Test pickling the scan data object."""
    obj1 = Scan((1, 2), (1, 1))
    obj2 = pickle_then_unpickle(obj1)
    assert obj1 == obj2



from __future__ import annotations

import sys

from dxtbx.format.FormatCBFMiniEiger import FormatCBFMiniEiger


class FormatCBFMiniEigerChessID7B2(FormatCBFMiniEiger):
    """A class for reading mini CBF format Eiger16M images for S/N E-32-0123
    installed at CHESS ID7B2, which has an inverted goniometer axis."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Eiger mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniEiger.get_cbf_header(image_file)
        for record in header.split("\n"):
            if "# Detector: Dectris EIGER2 Si 16M, S/N E-32-0123" in record:
                return True

        return False

    def _goniometer(self):
        return self._goniometer_factory.known_axis((-1, 0, 0))


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniEigerChessID7B2.understand(arg))
